# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from typing import Union

import montepy
from montepy.errors import *
from montepy.surfaces.surface_type import SurfaceType
from montepy.surfaces.surface import Surface, InitInput


class GeneralPlane(Surface):
    """Represents P

    .. versionchanged:: 1.0.0

        Added number parameter

    Parameters
    ----------
    input : Union[Input, str]
        The Input object representing the input
    number : int
        The number to set for this object.
    """

    def __init__(
        self,
        input: InitInput = None,
        number: int = None,
    ):
        super().__init__(input, number)
        if input:
            if self.surface_type != SurfaceType.P:
                raise ValueError("A GeneralPlane must be a surface of type P")
            if len(self.surface_constants) not in {4, 9}:
                raise ValueError(
                    "A GeneralPlane must have either 4 or 9 surface constants"
                )

    def validate(self):
        super().validate()
        if len(self.surface_constants) not in {4, 9}:
            raise IllegalState(
                f"Surface: {self.number} does not have constants set properly."
            )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.data_inputs.data_parser import parse_data
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input


from unittest import TestCase
import pytest


class TestTallyParser(TestCase):
    def test_parsing_tally_groups(self):
        for line in [
            "F4:n (1 3i 5) T",
            "F4:n (1 2 3 4 5) T",
            "F4:n 1 2 3",
            "F4:n (1 3i 5) (7 8 9) T",
            "f4:n (1 3i 5) (7 8 9)",
            "F7 (1 3i 5) (7 8 9)",
            "F7 (1 3i 5) (7 8 9) ",
        ]:
            print(line)
            input = Input([line], BlockType.DATA)
            data = parse_data(input)
            self.assertEqual(data.prefix, "f")

    def test_parsing_tally_print(self):
        input = Input(["Fq4 f p e"], BlockType.DATA)
        data = parse_data(input)
        self.assertEqual(data.prefix, "fq")

    def test_parsing_tally_multiplier(self):
        test_lines = {
            "fm904   (1.0) (1.0 961 103)",
            "fm3064 (1.0 361001 444) $ 1.0=mult",
        }
        for test in test_lines:
            print(test)
            input = Input([test], BlockType.DATA)
            data = parse_data(input)


@pytest.mark.parametrize(
    "line",
    [
        "fs14 -123",
        "fs12 -456 t",
        "fs11 -1 -2",
        "fs16 +1 +2 c",
        "fs17 -1 -2 t c",
    ],
)
def test_tally_segment_init(line):
    input = Input([line], BlockType.DATA)
    data = parse_data(input)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.

import montepy
from montepy.data_inputs import (
    data_input,
    fill,
    importance,
    lattice_input,
    material,
    mode,
    thermal_scattering,
    universe_input,
    volume,
)
from montepy.data_inputs import transform
import re

PREFIX_MATCHES = {
    fill.Fill,
    importance.Importance,
    lattice_input.LatticeInput,
    material.Material,
    mode.Mode,
    thermal_scattering.ThermalScatteringLaw,
    transform.Transform,
    volume.Volume,
    universe_input.UniverseInput,
}


def parse_data(input: montepy.mcnp_object.InitInput):
    """Parses the data input as the appropriate object if it is supported.

    Parameters
    ----------
    input : Union[Input, str]
        the Input object for this Data input

    Returns
    -------
    DataInput
        the parsed DataInput object
    """

    base_input = data_input.DataInput(input, fast_parse=True)
    prefix = base_input.prefix
    for data_class in PREFIX_MATCHES:
        if prefix == data_class._class_prefix():
            return data_class(input)
    return data_input.DataInput(input, prefix=prefix)



import argparse
import sys


def define_args(args):
    """Parses the arguments from the command line.

    Parameters
    ----------
    args : list
        the arguments from the command line.

    Returns
    -------
    argparse.Namespace
        the parsed arguments (with argparse)
    """
    parser = argparse.ArgumentParser(
        prog="Change_to_ascii",
        description="Change the encoding of a file to strict ASCII. Everything not compliant will be removed.",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-d",
        "--delete",
        dest="delete",
        action="store_true",
        help="Delete any non-ascii characters. This is the default.",
    )
    group.add_argument(
        "-w",
        "--whitespace",
        dest="whitespace",
        action="store_true",
        help="Replace non-ascii characters with a space.",
    )
    parser.add_argument("in_file", nargs=1, help="The input file to convert")
    parser.add_argument("out_file", nargs=1, help="The input file to convert")
    args = parser.parse_args(args)
    return args


def strip_characters(args):
    """Strips non-ascii characters from the input file, and writes out the output file.

    Parameters
    ----------
    args : argparse.Namespace
        the parsed command line arguments.
    """
    if args.whitespace:
        replacer = " "
    elif args.delete:
        replacer = ""
    # default to delete
    else:
        replacer = ""
    with open(args.in_file[0], "rb") as in_fh, open(args.out_file[0], "wb") as out_fh:
        for line in in_fh:
            utf8_line = line.decode(encoding="utf8", errors="replace")
            utf8_line = utf8_line.replace("�", replacer)

            try:
                out_fh.write(utf8_line.encode(encoding="ascii", errors="strict"))
            except UnicodeError as e:
                new_line = []
                # find the bad characters character by character
                for char in utf8_line:
                    if ord(char) > 128:
                        new_line.append(replacer)
                    else:
                        new_line.append(char)
                out_fh.write(
                    "".join(new_line).encode(encoding="ascii", errors="strict")
                )


def main(args=None):
    """Main runner function.

    Parameters
    ----------
    args : list
        The arguments passed from the command line.
    """
    if args is None:
        args = sys.argv[1:]
    args = define_args(args)
    strip_characters(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import montepy
from montepy.data_inputs.data_parser import parse_data
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input

import pytest


@pytest.mark.parametrize(
    "line", ["sdef  cel=d1 erg=d2 pos=fcel d3 ext=fcel d4 axs=0 0 1 rad=d5"]
)
def test_source_parse_and_parrot(line):
    input = Input([line], BlockType.DATA)
    data = parse_data(input)
    print(repr(data._tree))
    assert data._tree.format() == line



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.constants import DEFAULT_VERSION


def read_input(destination, mcnp_version=DEFAULT_VERSION, replace=True):
    """Reads the specified MCNP Input file.

    The MCNP version must be a three component tuple e.g., (6, 2, 0) and (5, 1, 60).

    Notes
    -----
    if a stream is provided. It will not be closed by this function.

    Parameters
    ----------
    destination : io.TextIOBase, str, os.PathLike
        the path to the input file to read, or a readable stream.
    mcnp_version : tuple
        The version of MCNP that the input is intended for.
    replace : bool
        replace all non-ASCII characters with a space (0x20)

    Returns
    -------
    MCNP_Problem
        The MCNP_Problem instance representing this file.

    Raises
    ------
    UnsupportedFeature
        If an input format is used that MontePy does not support.
    MalformedInputError
        If an input has a broken syntax.
    NumberConflictError
        If two objects use the same number in the input file.
    BrokenObjectLinkError
        If a reference is made to an object that is not in the input
        file.
    UnknownElement
        If an isotope is specified for an unknown element.
    """
    problem = montepy.mcnp_problem.MCNP_Problem(destination)
    problem.mcnp_version = mcnp_version
    problem.parse_input(replace=replace)
    return problem



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.errors import *
from montepy.input_parser.tokens import DataLexer
from montepy.input_parser.parser_base import MCNP_Parser, MetaBuilder
from montepy.input_parser import syntax_node


class DataParser(MCNP_Parser):
    """A parser for almost all data inputs.

    Returns
    -------
    SyntaxNode
        a syntax tree for the data input.
    """

    debugfile = None

    @_(
        "introduction",
        "introduction data",
        "introduction data parameters",
        "introduction parameters",
    )
    def data_input(self, p):
        ret = {}
        for key, node in p.introduction.nodes.items():
            ret[key] = node
        if hasattr(p, "data"):
            ret["data"] = p.data
        else:
            ret["data"] = syntax_node.ListNode("empty data")
        if hasattr(p, "parameters"):
            ret["parameters"] = p.parameters
        return syntax_node.SyntaxNode("data", ret)

    @_(
        "classifier_phrase",
        "classifier_phrase MODIFIER padding",
        "padding classifier_phrase",
        "padding classifier_phrase MODIFIER padding",
    )
    def introduction(self, p):
        ret = {}
        if isinstance(p[0], syntax_node.PaddingNode):
            ret["start_pad"] = p[0]
        else:
            ret["start_pad"] = syntax_node.PaddingNode()
        ret["classifier"] = p.classifier_phrase
        if hasattr(p, "MODIFIER"):
            ret["keyword"] = syntax_node.ValueNode(p.MODIFIER, str, padding=p[-1])
        else:
            ret["keyword"] = syntax_node.ValueNode(None, str, padding=None)
        return syntax_node.SyntaxNode("data intro", ret)

    @_(
        "number_sequence",
        "isotope_fractions",
        "particle_sequence",
        "text_sequence",
        "kitchen_sink",
    )
    def data(self, p):
        return p[0]

    @_("zaid_phrase number_phrase")
    def isotope_fraction(self, p):
        return p

    @_("isotope_fraction", "isotope_fractions isotope_fraction")
    def isotope_fractions(self, p):
        if hasattr(p, "isotope_fractions"):
            fractions = p.isotope_fractions
        else:
            fractions = syntax_node.MaterialsNode("isotope list")
        fractions.append_nuclide(p.isotope_fraction)
        return fractions

    @_("ZAID", "ZAID padding")
    def zaid_phrase(self, p):
        return self._flush_phrase(p, str)

    @_("particle_phrase", "particle_sequence particle_phrase")
    def particle_sequence(self, p):
        if len(p) == 1:
            sequence = syntax_node.ListNode("particle sequence")
            sequence.append(p[0], True)
        else:
            sequence = p[0]
            sequence.append(p[1], True)
        return sequence

    @_("PARTICLE", "SURFACE_TYPE", "PARTICLE_SPECIAL")
    def particle_text(self, p):
        return p[0]

    @_("particle_text padding", "particle_text")
    def particle_phrase(self, p):
        return self._flush_phrase(p, str)

    # Manually specifying because more levels break SLY. Might be hitting some hard coded limit.
    @_(
        "NUMBER_WORD",
        "NUM_MULTIPLY",
        "NUMBER_WORD padding ",
        "NUM_MULTIPLY padding",
    )
    def text_phrase(self, p):
        self._flush_phrase(p, str)

    @_("text_phrase", "text_sequence text_phrase")
    def text_sequence(self, p):
        if len(p) == 1:
            sequence = syntax_node.ListNode("text sequence")
            sequence.append(p[0], True)
        else:
            sequence = p[0]
            sequence.append(p[1], True)
        return sequence

    @_("kitchen_junk", "kitchen_sink kitchen_junk")
    def kitchen_sink(self, p):
        sequence = p[0]
        if len(p) != 1:
            for node in p[1].nodes:
                sequence.append(node, True)
        return sequence

    @_("number_sequence", "text_sequence", "particle_sequence")
    def kitchen_junk(self, p):
        return p[0]

    @_("classifier param_seperator NUMBER text_phrase")
    def parameter(self, p):
        return syntax_node.SyntaxNode(
            p.classifier.prefix.value,
            {
                "classifier": p.classifier,
                "seperator": p.param_seperator,
                "data": syntax_node.ValueNode(p.NUMBER + p.text_phrase, str),
            },
        )


class ClassifierParser(DataParser):
    """A parser for parsing the first word or classifier of a data input.

    Returns
    -------
    ClassifierNode
        the classifier of the data input.
    """

    debugfile = None

    @_("classifier", "padding classifier")
    def data_classifier(self, p):
        if hasattr(p, "padding"):
            padding = p.padding
        else:
            padding = syntax_node.PaddingNode(None)
        return syntax_node.SyntaxNode(
            "data input classifier", {"start_pad": padding, "classifier": p.classifier}
        )


class ParamOnlyDataParser(DataParser):
    """A parser for parsing parameter (key-value pair) only data inputs.

    .e.g., SDEF

    .. versionadded:: 0.3.0

    Returns
    -------
    SyntaxNode
        a syntax tree for the data input.
    """

    debugfile = None

    @_(
        "param_introduction",
        "param_introduction spec_parameters",
    )
    def param_data_input(self, p):
        ret = {}
        for key, node in p.param_introduction.nodes.items():
            ret[key] = node
        if hasattr(p, "spec_parameters"):
            ret["parameters"] = p.spec_parameters
        return syntax_node.SyntaxNode("data", ret)

    @_(
        "classifier_phrase",
        "padding classifier_phrase",
    )
    def param_introduction(self, p):
        ret = {}
        if isinstance(p[0], syntax_node.PaddingNode):
            ret["start_pad"] = p[0]
        else:
            ret["start_pad"] = syntax_node.PaddingNode()
        ret["classifier"] = p.classifier_phrase
        ret["keyword"] = syntax_node.ValueNode(None, str, padding=None)
        return syntax_node.SyntaxNode("data intro", ret)

    @_("spec_parameter", "spec_parameters spec_parameter")
    def spec_parameters(self, p):
        """A list of the parameters (key, value pairs) for this input.

        Returns
        -------
        ParametersNode
            all parameters
        """
        if len(p) == 1:
            params = syntax_node.ParametersNode()
            param = p[0]
        else:
            params = p[0]
            param = p[1]
        params.append(param)
        return params

    @_("spec_classifier param_seperator data")
    def spec_parameter(self, p):
        return syntax_node.SyntaxNode(
            p.spec_classifier.prefix.value,
            {
                "classifier": p.spec_classifier,
                "seperator": p.param_seperator,
                "data": p.data,
            },
        )

    @_(
        "KEYWORD",
    )
    def spec_data_prefix(self, p):
        return syntax_node.ValueNode(p[0], str)

    @_(
        "modifier spec_data_prefix",
        "spec_data_prefix",
        "spec_classifier NUMBER",
        "spec_classifier particle_type",
    )
    def spec_classifier(self, p):
        """The classifier of a data input.

        This represents the first word of the data input.
        E.g.: ``M4``, `IMP:N`, ``F104:p``

        Returns
        -------
        ClassifierNode
        """
        if hasattr(p, "spec_classifier"):
            classifier = p.spec_classifier
        else:
            classifier = syntax_node.ClassifierNode()

        if hasattr(p, "modifier"):
            classifier.modifier = syntax_node.ValueNode(p.modifier, str)
        if hasattr(p, "spec_data_prefix"):
            classifier.prefix = p.spec_data_prefix
        if hasattr(p, "NUMBER"):
            classifier.number = syntax_node.ValueNode(p.NUMBER, int)
        if hasattr(p, "particle_type"):
            classifier.particles = p.particle_type
        return classifier



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
import montepy
from montepy.numbered_object_collection import NumberedDataObjectCollection
from montepy.data_inputs.transform import Transform


class Transforms(NumberedDataObjectCollection):
    """A container of multiple :class:`~montepy.data_inputs.transform.Transform` instances.

    Notes
    -----

    For examples see the ``NumberedObjectCollection`` :ref:`collect ex`.
    """

    def __init__(self, objects: list = None, problem: montepy.MCNP_Problem = None):
        super().__init__(Transform, objects, problem)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.input_parser.data_parser import DataParser
from montepy.input_parser import syntax_node


class TallySegmentParser(DataParser):
    """A barebone parser for parsing tally segment inputs before they are fully implemented.

    .. versionadded:: 0.2.10

    Returns
    -------
    SyntaxNode
        a syntax tree for the data input.
    """

    debugfile = None

    @_("introduction tally_specification")
    def tally(self, p):
        ret = {}
        for key, node in p.introduction.nodes.items():
            ret[key] = node
        ret["tally"] = p.tally_specification
        return syntax_node.SyntaxNode("data", ret)

    @_(
        "tally_numbers",
        "tally_numbers end_phrase",
        "tally_numbers end_phrase end_phrase",
    )
    def tally_specification(self, p):
        if hasattr(p, "end_phrase"):
            text = p.end_phrase
        else:
            text = syntax_node.ValueNode(None, str)

        return syntax_node.SyntaxNode(
            "tally list", {"tally": p.tally_numbers, "end": text}
        )

    @_("PARTICLE", "PARTICLE padding", "TEXT", "TEXT padding")
    def end_phrase(self, p):
        """A non-zero number with or without padding.

        Returns
        -------
        ValueNode
            a float ValueNode
        """
        return self._flush_phrase(p, str)

    @_(
        "tally_numbers tally_numbers",
        "number_sequence",
        "tally_numbers padding",
    )
    def tally_numbers(self, p):
        if hasattr(p, "tally_numbers"):
            ret = p.tally_numbers
            ret.nodes["right"] += p.padding
            return ret
        if hasattr(p, "tally_numbers1"):
            return syntax_node.SyntaxNode("tally tree", {"left": p[0], "right": p[1]})
        else:
            left = syntax_node.PaddingNode(None)
            right = syntax_node.PaddingNode(None)
            return syntax_node.SyntaxNode(
                "tally set", {"left": left, "tally": p[0], "right": right}
            )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
import montepy
from montepy.surfaces.surface import Surface
from montepy.surfaces.surface_type import SurfaceType
from montepy.numbered_object_collection import NumberedObjectCollection


def __create_surface_generator(surf_type):
    """A function for a Descriptor for the Surfaces collection.

    This is meant to make it possible to get a generator at
    surfaces.pz

    This works by creating a closure that is a generator.
    This closure is then passed to ``property()`` to create a property
    """

    def closure(obj):
        for surf in obj.objects:
            if surf.surface_type == surf_type:
                yield surf

    return closure


class Surfaces(NumberedObjectCollection):
    """A collection of multiple :class:`montepy.surfaces.surface.Surface` instances.

    This collection has a generator for every supported type of MCNP surface.
    These are accessed by the for lower case version of the MCNP mnemonic.
    For example you can access all planes normal to the z-axis through .pz

    This example will shift all PZ surfaces up by 10 cm.

    .. testcode:: python

        import montepy
        problem = montepy.read_input("tests/inputs/test.imcnp")
        for surface in problem.surfaces.pz:
            surface.location += 10

    Notes
    -----

    For examples see the ``NumberedObjectCollection`` :ref:`collect ex`.

    Parameters
    ----------
    surfaces : list
        the list of surfaces to start with if needed
    """

    def __init__(self, surfaces: list = None, problem: montepy.MCNP_Problem = None):
        super().__init__(Surface, surfaces, problem)


def __setup_surfaces_generators():
    for surf_type in SurfaceType:
        doc = f"Generator for getting all surfaces of type *{surf_type.description}* or ``{surf_type.value}``"
        getter = property(__create_surface_generator(surf_type), doc=doc)
        setattr(Surfaces, surf_type.name.lower(), getter)


__setup_surfaces_generators()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools as it
from montepy.data_inputs.cell_modifier import CellModifierInput, InitInput
from montepy.data_inputs.transform import Transform
from montepy.errors import *
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input, Jump
from montepy.input_parser import syntax_node
from montepy.mcnp_object import MCNP_Object
from montepy.universe import Universe
from montepy.utilities import *
import numpy as np


class Fill(CellModifierInput):
    """Object to handle the ``FILL`` input in cell and data blocks.

    Parameters
    ----------
    input : Union[Input, str]
        the Input object representing this data input
    in_cell_block : bool
        if this card came from the cell block of an input file.
    key : str
        the key from the key-value pair in a cell
    value : SyntaxNode
        the value syntax tree from the key-value pair in a cell
    """

    DIMENSIONS = {"i": 0, "j": 1, "k": 2}
    """Maps the dimension to its axis number"""

    def __init__(
        self,
        input: InitInput = None,
        in_cell_block: bool = False,
        key: str = None,
        value: syntax_node.SyntaxNode = None,
    ):
        self._old_number = self._generate_default_node(int, None)
        self._old_numbers = None
        self._universe = None
        self._universes = None
        self._transform = None
        self._hidden_transform = None
        self._old_transform_number = None
        self._multi_universe = False
        super().__init__(input, in_cell_block, key, value)
        if self.in_cell_block:
            if key:
                self._parse_cell_input(key, value)
        elif input:
            self._old_numbers = []
            values = self.data
            for value in values:
                try:
                    value._convert_to_int()
                    if value.value is not None:
                        assert value.value >= 0
                        self._old_numbers.append(value)
                    else:
                        self._old_numbers.append(value)
                except (ValueError, AssertionError) as e:
                    raise MalformedInputError(
                        input,
                        f"Cell fill must be set to a valid universe, {value} was given",
                    )

    def _generate_default_cell_tree(self):
        list_node = syntax_node.ListNode("number sequence")
        list_node.append(self._generate_default_node(float, None))
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = self._generate_default_node(
            str, self._class_prefix().upper(), None
        )
        self._tree = syntax_node.SyntaxNode(
            "fill",
            {
                "classifier": classifier,
                "param_seperator": self._generate_default_node(str, "=", None),
                "data": list_node,
            },
        )

    def _parse_cell_input(self, key, value):
        """Parses the information provided in the cell input.

        Parameters
        ----------
        key : str
            The key given in the cell
        value : str
            the value given in the cell
        """

        def get_universe(value):
            if ":" in value["data"].nodes:
                self._parse_matrix(value)
            else:
                data = value["data"]
                try:
                    val = data[0]
                    val._convert_to_int()
                    assert val.value >= 0
                    self._old_number = val
                except (TypeError, AssertionError) as e:
                    raise ValueError(
                        f"The fill universe must be a valid integer ≥ 0, {data} was given"
                    )
                # ensure only one universe is given
                if (
                    len(data) >= 2
                    and isinstance(data[1], syntax_node.ValueNode)
                    and "(" != data[1].value
                ):
                    raise ValueError(
                        f"Fill cannot have two universes in this format. {data.format()} given"
                    )

        data = value["data"]
        if not isinstance(data, syntax_node.ListNode):
            data = data.nodes
        if "(" in data:
            get_universe(value)
            trans_data = value["data"][
                list(value["data"]).index("(") + 1 : list(value["data"]).index(")") - 1
            ]
            if len(trans_data) == 1:
                try:
                    transform = trans_data[0]
                    transform._convert_to_int()
                    assert transform.value > 0
                    self._hidden_transform = False
                    self._old_transform_number = transform
                except AssertionError as e:
                    raise ValueError(
                        "Transform number must be a positive integer. {words[0]} was given."
                    )
            elif len(trans_data) > 1:
                modifier = value["classifier"].modifier
                if modifier and "*" in modifier.value:
                    in_key = "*TR1"
                else:
                    in_key = "TR1"
                input_card = Input([in_key + " " + trans_data.format()], BlockType.DATA)
                self._transform = Transform(input_card, pass_through=True)
                self._hidden_transform = True

        else:
            get_universe(value)

    def _parse_matrix(self, value):
        """Parses a matrix fill of universes.

        Parameters
        ----------
        value : str
            the value in the cell
        """
        self._multi_universe = True
        words = value["data"]
        self._min_index = np.zeros((3,), dtype=np.dtype(int))
        self._max_index = np.zeros((3,), dtype=np.dtype(int))
        limits_iter = (
            it.islice(words, 0, None, 3),
            it.islice(words, 1, None, 3),
            it.islice(words, 2, None, 3),
        )
        for axis, min_val, seperator, max_val in zip(
            Fill.DIMENSIONS.values(), *limits_iter
        ):
            for val, limit_holder in zip(
                (min_val, max_val), (self._min_index, self._max_index)
            ):
                try:
                    val._convert_to_int()
                    limit_holder[axis] = val.value
                except ValueError as e:
                    raise ValueError(
                        f"The lattice limits must be an integer. {val.value} was given"
                    )
        for min_val, max_val in zip(self.min_index, self.max_index):
            if min_val > max_val:
                raise ValueError(
                    "The minimum value must be smaller than the max value."
                    f"Min: {min_val}, Max: {max_val}, Input: {value.format()}"
                )
        self._old_numbers = np.zeros(self._sizes, dtype=np.dtype(int))
        words = iter(words[9:])
        for i in self._axis_range(0):
            for j in self._axis_range(1):
                for k in self._axis_range(2):
                    val = next(words)
                    try:
                        val._convert_to_int()
                        assert val.value >= 0
                        self._old_numbers[i][j][k] = val.value
                    except (ValueError, AssertionError) as e:
                        raise ValueError(
                            f"Values provided must be valid universes. {val.value} given."
                        )

    @staticmethod
    def _class_prefix():
        return "fill"

    @staticmethod
    def _has_number():
        return False

    @staticmethod
    def _has_classifier():
        return 0

    @property
    def universe(self):
        """The universe that this cell will be filled with.

        Only returns a value when :func:`multiple_universes` is False, otherwise none.

        Returns
        -------
        Universe
            the universe that the cell will be filled with, or None
        """
        if not self.multiple_universes:
            return self._universe

    @universe.setter
    def universe(self, value):
        if not isinstance(value, (Universe, type(None))):
            raise TypeError(f"Universe must be set to a Universe. {value} given.")
        if self.multiple_universes:
            raise ValueError(
                "A single universe can only be set when multiple_universes is False."
            )
        self._universe = value

    @universe.deleter
    def universe(self):
        self._universe = None

    @property
    def universes(self):
        """The universes that this cell will be filled with in a lattice.

        Only returns a value when :func:`multiple_universes` is true, otherwise none.

        Returns
        -------
        np.ndarray
            the universes that the cell will be filled with as a 3-D
            array.
        """
        if self.multiple_universes:
            return self._universes

    @universes.setter
    def universes(self, value):
        if not isinstance(value, (np.ndarray, type(None))):
            raise TypeError(f"Universes must be set to an array. {value} given.")
        if not self.multiple_universes:
            raise ValueError(
                "Multiple universes can only be set when multiple_universes is True."
            )
        self._universes = value

    @universes.deleter
    def universes(self):
        self._universes = None

    @property
    def min_index(self):
        """The minimum indices of the matrix in each dimension.

        For the order of the indices see: ``DIMENSIONS``.

        Returns
        -------
        :class:`numpy.ndarry`
            the minimum indices of the matrix for complex fills
        """
        return self._min_index

    @property
    def max_index(self):
        """The maximum indices of the matrix in each dimension.

        For the order of the indices see: ``DIMENSIONS``.

        Returns
        -------
        :class:`numpy.ndarry`
            the maximum indices of the matrix for complex fills
        """
        return self._max_index

    @property
    def multiple_universes(self):
        """Whether or not this cell is filled with multiple universes in a matrix.

        Returns
        -------
        bool
            True if this cell contains multiple universes
        """
        return self._multi_universe

    @multiple_universes.setter
    def multiple_universes(self, value):
        if not isinstance(value, bool):
            raise TypeError("Multiple_univeses must be set to a bool")
        self._multi_universe = value

    @make_prop_val_node("_old_number")
    def old_universe_number(self):
        """The number of the universe that this is filled by taken from the input.

        Returns
        -------
        int
            the old universe number
        """
        pass

    @property
    def old_universe_numbers(self):
        """The numbers of the universes that this is filled by taken from the input.

        Returns
        -------
        :class:`numpy.ndarray`
            the old universe numbers
        """
        if isinstance(self._old_numbers, list):
            return [
                num.value if isinstance(num, syntax_node.ValueNode) else num
                for num in self._old_numbers
            ]
        return self._old_numbers

    @property
    def hidden_transform(self):
        """Whether or not the transform used is hidden.

        This is true when an unnumbered transform is used
        e.g., ``FILL=1 (1.0 2.0 3.0)``.

        Returns
        -------
        bool
            True iff the transform used is hidden
        """
        return self._hidden_transform

    @property
    def has_information(self):
        if self.in_cell_block:
            return self.universe is not None or self.universes is not None

    @property
    def _tree_value(self):
        if self.transform or self.multiple_universes:
            raise ValueError(
                f"Fill can not be in the data block if"
                " fill transforms and other complex inputs are used."
            )
        val = self._old_number
        val.value = self.universe.number if self.universe else None
        return val

    @property
    def transform(self):
        """The transform for this fill (if any).

        Returns
        -------
        Transform
            the transform for the filling universe for this cell.
        """
        return self._transform

    @transform.setter
    def transform(self, value):
        if not isinstance(value, (Transform, type(None))):
            raise TypeError("Transform must be set to a Transform.")
        self._transform = value
        if value is not None:
            self._hidden_transform = value.hidden_transform
        else:
            self._hidden_transform = False

    @transform.deleter
    def transform(self):
        self._transform = None

    @make_prop_val_node("_old_transform_number")
    def old_transform_number(self):
        """The number of the transform specified in the input.

        Returns
        -------
        int
            the original number for the transform from the input.
        """
        pass

    def merge(self, other):
        raise MalformedInputError(
            other._input, "Cannot have two lattice inputs for the problem"
        )

    def push_to_cells(self):
        def get_universe(number):
            return self._problem.universes[number]

        if self.in_cell_block:
            if self.old_transform_number:
                self._transform = self._problem.transforms[self.old_transform_number]
            if (
                self.old_universe_number is not None
                or self.old_universe_numbers is not None
            ):
                if isinstance(self.old_universe_numbers, np.ndarray):
                    self._universes = np.empty_like(
                        self.old_universe_numbers, dtype="O"
                    )
                    for i in self._axis_range(0):
                        for j in self._axis_range(1):
                            for k in self._axis_range(2):
                                self._universes[i][j][k] = get_universe(
                                    self.old_universe_numbers[i][j][k].item()
                                )
                else:
                    self._universe = get_universe(self.old_universe_number)
        else:
            if not self.set_in_cell_block and self.old_universe_numbers:
                for cell, old_number in zip(self._problem.cells, self._old_numbers):
                    if not isinstance(old_number, Jump):
                        cell._fill._old_number = old_number
            for cell in self._problem.cells:
                cell._fill.push_to_cells()

    def _clear_data(self):
        self._old_number = None
        self._universe = None

    def _axis_range(self, axis):
        """Returns an iterator for iterating over the given axis.

        Parameters
        ----------
        axis : int
            the number of the axis to iterate over

        Returns
        -------
        unknown
            range
        """
        return range(self._axis_size(axis))

    def _axis_size(self, axis):
        """Get the length of the given axis.

        Parameters
        ----------
        axis : int
            the axis to probe into.

        Returns
        -------
        int
            the length of the given axis of the universe matrix.
        """
        return int(self.max_index[axis] - self.min_index[axis]) + 1

    @property
    def _sizes(self):
        """The axis sizes of the matrix.

        Returns
        -------
        tuple
            a tuple of the matrix shape.
        """
        return (self._axis_size(0), self._axis_size(1), self._axis_size(2))

    def __str__(self):
        return f"Fill: Universe: {self.universe}, transform: {self.transform}"

    def __repr__(self):
        return (
            f"Fill: set_in_cell: {self.set_in_cell_block}, in_cell: {self.in_cell_block} "
            f"old_number: {self.old_universe_number}, old_transform: {self._old_transform_number} "
            f"old_numbers: {self.old_universe_numbers} "
            f"Universe: {self.universe}, universes: {self.universes}, transform: {self.transform} "
            f"Multi_universe: {self._multi_universe} hidden_transform: {self.hidden_transform} "
            f"Min/Max: {str(self.min_index) + ' ' +str(self.max_index) if self._multi_universe == True  else 'None'}"
        )

    def _update_cell_values(self):
        # Todo update matrix fills
        new_vals = list(self._tree["data"])
        if self.transform and self.transform.is_in_degrees:
            self._tree["classifier"].modifier = "*"
        else:
            self._tree["classifier"].modifier = None
        new_vals = self._update_cell_universes(new_vals)
        if self.transform is None:
            try:
                values = [val.value for val in self._tree["data"]]
                start = values.index("(")
                end = values.index(")")
                del new_vals[start : end + 1]
            except ValueError:
                pass
        # Update transforms
        else:
            start = -1
            end = -1
            try:
                values = [val.value for val in self._tree["data"]]
                start = values.index("(")
                end = values.index(")")
            except ValueError:
                pass
            if self.transform.hidden_transform:
                self.transform._update_values()
                payload = list(self.transform._tree["data"])
            else:
                # if started with named transform
                if start > 0 and end > 0 and ((end - start) - 1 == 1):
                    payload = [self._tree["data"][start + 1]]
                else:
                    payload = [syntax_node.ValueNode("1", int)]
                payload[0].value = self.transform.number
            if start > 0 and end > 0:
                new_vals = new_vals[: start + 1] + payload + new_vals[end:]
        self._tree["data"].update_with_new_values(new_vals)

    def _update_cell_universes(self, new_vals):
        def _value_node_generator():
            while True:
                value = syntax_node.ValueNode("1", int)
                padding = syntax_node.PaddingNode(" ")
                value.padding = padding
                yield value

        if self.multiple_universes:
            payload = []
            for i in self._axis_range(0):
                for j in self._axis_range(1):
                    for k in self._axis_range(2):
                        payload.append(self.universes[i][j][k].number)
        else:
            payload = [
                (
                    self.universe.number
                    if self.universe is not None
                    else self.old_universe_number
                )
            ]
        try:
            start_transform = new_vals.index("(")
        except ValueError:
            start_transform = None

        reverse_list = new_vals.copy()
        reverse_list.reverse()
        try:
            start_matrix = len(new_vals) - reverse_list.index(":") + 1
        except ValueError:
            start_matrix = 0
        value_nodes = it.chain(
            new_vals[start_matrix:start_transform], _value_node_generator()
        )
        buffer = []
        for universe, value in zip(payload, value_nodes):
            value.value = universe
            buffer.append(value)
        buffer = new_vals[:start_matrix] + buffer
        if start_transform:
            buffer += new_vals[start_transform:]
        return buffer



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from hypothesis import assume, given, note, strategies as st, settings, HealthCheck
import pathlib
import pytest

import montepy

from montepy.data_inputs.element import Element
from montepy.data_inputs.nuclide import Nucleus, Nuclide, Library
from montepy.data_inputs.material import Material, _DefaultLibraries as DL
from montepy.data_inputs.material_component import MaterialComponent
from montepy.data_inputs.thermal_scattering import ThermalScatteringLaw
from montepy.errors import MalformedInputError, UnknownElement
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input
from montepy.particle import LibraryType


# test material
class TestMaterial:
    @pytest.fixture
    def big_material(_):
        components = [
            "h1.00c",
            "h1.04c",
            "h1.80c",
            "h1.04p",
            "h2",
            "h3",
            "th232",
            "th232.701nc",
            "U235",
            "U235.80c",
            "U235m1.80c",
            "u238",
            "am242",
            "am242m1",
            "Co60m2.50c",
            "Pu239",
            "Ni-0.60c",
        ]
        mat = Material()
        mat.number = 1
        for component in components:
            mat.add_nuclide(component, 0.05)
        return mat

    @pytest.fixture
    def parsed_material(_):
        return Material(
            """M1   1001.00c   0.05
      1001.04c   0.05
      1001.80c   0.05
      1001.04p   0.05
      1002       0.05
      1003       0.05
      90232      0.05
      90232.701nc 0.05
      92235      0.05
      92235.80c  0.05
      92635.80c  0.05
      92238      0.05
      95642      0.05
      95242      0.05
      27560.50c  0.05
      94239      0.05
      28000.60c  0.05"""
        )

    @pytest.fixture
    def materials(_, big_material, parsed_material):
        return (big_material, parsed_material)

    @pytest.fixture
    def big_mat_lib(_, big_material):
        mat = big_material
        mat.default_libraries["nlib"] = "00c"
        mat.default_libraries["plib"] = "80p"
        return mat

    @pytest.fixture
    def prob_default(_):
        prob = montepy.MCNP_Problem("hi")
        prob.materials.default_libraries["alib"] = "24a"
        return prob

    @pytest.mark.parametrize(
        "isotope_str, lib_type, lib_str",
        [
            ("H-1.80c", "nlib", "80c"),
            ("H-1.80c", "plib", "80p"),
            ("H-1.80c", "hlib", None),
            ("H-1.80c", "alib", "24a"),
        ],
    )
    def test_mat_get_nuclide_library(
        _, big_mat_lib, prob_default, isotope_str, lib_type, lib_str
    ):
        nuclide = Nuclide(isotope_str)
        if lib_str:
            lib = Library(lib_str)
            big_mat_lib.link_to_problem(prob_default)
        else:
            lib = None
        assert big_mat_lib.get_nuclide_library(nuclide, lib_type) == lib
        assert (
            big_mat_lib.get_nuclide_library(nuclide, LibraryType(lib_type.upper()))
            == lib
        )
        if lib is None:
            big_mat_lib.link_to_problem(prob_default)
            assert big_mat_lib.get_nuclide_library(nuclide, lib_type) == lib
        # test iter, items defaults
        for iter_key, (item_key, item_val) in zip(
            big_mat_lib.default_libraries, big_mat_lib.default_libraries.items()
        ):
            assert iter_key == item_key
            assert big_mat_lib.default_libraries[iter_key] == item_val

    def test_mat_clear(_, big_mat_lib):
        old_mat = big_mat_lib.clone()
        old_len = len(big_mat_lib)
        assert old_len > 0
        big_mat_lib.clear()
        assert 0 == len(big_mat_lib)
        for part in old_mat.default_libraries:
            assert (
                old_mat.default_libraries[part] == big_mat_lib.default_libraries[part]
            )
        big_mat_lib.add_nuclide("O-16.80c", 1.0)
        _.verify_export(big_mat_lib)

    def test_mat_get_nuclide_library_bad(_, big_mat_lib):
        with pytest.raises(TypeError):
            big_mat_lib.get_nuclide_library(5, "nlib")
        with pytest.raises(TypeError):
            big_mat_lib.get_nuclide_library("1001.80c", 5)

    def test_material_parameter_parsing(_):
        for line in [
            "M20 1001.80c 1.0 gas=0",
            "M20 1001.80c 1.0 gas = 0 nlib = 00c",
            "M120 nlib=80c 1001 1.0",
        ]:
            material = Material(line)

    def test_material_validator(_):
        material = Material()
        with pytest.raises(montepy.errors.IllegalState):
            material.validate()
        with pytest.raises(montepy.errors.IllegalState):
            material.format_for_mcnp_input((6, 2, 0))

    def test_material_number_setter(_):
        in_str = "M20 1001.80c  0.5  8016.80c  0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        material.number = 30
        assert material.number == 30
        with pytest.raises(TypeError):
            material.number = "foo"
        with pytest.raises(ValueError):
            material.number = -5
        _.verify_export(material)

    @pytest.mark.filterwarnings("ignore")
    def test_material_is_atom_frac_setter(_, big_material):
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input = Input([in_str], BlockType.DATA)
        material = Material(input)
        assert material.is_atom_fraction
        _.verify_export(material)
        material.is_atom_fraction = False
        assert not material.is_atom_fraction
        _.verify_export(material)
        for frac_type in [False, True]:
            big_material.is_atom_fraction = frac_type
            assert big_material.is_atom_fraction == frac_type
            _.verify_export(big_material)

    def test_material_getter_iter(_, big_material):
        for i, (nuclide, frac) in enumerate(big_material):
            gotten = big_material[i]
            assert gotten[0] == nuclide
            assert gotten[1] == pytest.approx(frac)
        comp_0, comp_1 = big_material[0:2]
        assert comp_0 == big_material[0]
        assert comp_1 == big_material[1]
        _, comp_1 = big_material[0:4:3]
        assert comp_1 == big_material[3]
        with pytest.raises(TypeError):
            big_material["hi"]

    def test_material_setter(_, big_material):
        big_material[2] = (Nuclide("1001.80c"), 1.0)
        assert big_material[2][0] == Nuclide("1001.80c")
        assert big_material[2][1] == pytest.approx(1.0)
        with pytest.raises(TypeError):
            big_material["hi"] = 5
        with pytest.raises(TypeError):
            big_material[2] = 5
        with pytest.raises(ValueError):
            big_material[2] = (5,)
        with pytest.raises(TypeError):
            big_material[2] = (5, 1.0)
        with pytest.raises(TypeError):
            big_material[2] = (Nuclide("1001.80c"), "hi")
        with pytest.raises(ValueError):
            big_material[2] = (Nuclide("1001.80c"), -1.0)
        _.verify_export(big_material)

    def test_material_deleter(_, big_material):
        old_comp = big_material[6]
        del big_material[6]
        assert old_comp[0] not in big_material.nuclides
        old_comps = big_material[0:2]
        del big_material[0:2]
        for nuc, _f in old_comps:
            assert nuc not in big_material.nuclides
        with pytest.raises(TypeError):
            del big_material["hi"]
        pu_comp = big_material[-1]
        del big_material[-1]
        assert pu_comp[0] not in big_material.nuclides
        _.verify_export(big_material)

    def test_material_values(_, materials):
        for big_material in materials:
            # test iter
            for value in big_material.values:
                assert value == pytest.approx(0.05)
            assert len(list(big_material.values)) == len(big_material)
            # test getter setter
            for i, comp in enumerate(big_material):
                assert big_material.values[i] == pytest.approx(comp[1])
                big_material.values[i] = 1.0
                assert big_material[i][1] == pytest.approx(1.0)
            with pytest.raises(TypeError):
                big_material.values["hi"]
            with pytest.raises(IndexError):
                big_material.values[len(big_material) + 1]
            with pytest.raises(TypeError):
                big_material.values[0] = "hi"
            with pytest.raises(ValueError):
                big_material.values[0] = -1.0
            _.verify_export(big_material)

    def test_material_nuclides(_, materials):
        for big_material in materials:
            # test iter
            for nuclide, comp in zip(big_material.nuclides, big_material):
                assert nuclide == comp[0]
                # test getter setter
            for i, comp in enumerate(big_material):
                assert big_material.nuclides[i] == comp[0]
                big_material.nuclides[i] = Nuclide("1001.80c")
                assert big_material[i][0] == Nuclide("1001.80c")
            with pytest.raises(TypeError):
                big_material.nuclides["hi"]
            with pytest.raises(IndexError):
                big_material.nuclides[len(big_material) + 1]
            with pytest.raises(TypeError):
                big_material.nuclides[0] = "hi"
            _.verify_export(big_material)

    @given(st.integers(1, 99), st.floats(1.9, 2.3), st.floats(0, 20, allow_nan=False))
    def test_material_append(_, Z, a_multiplier, fraction):
        mat = Material()
        mat.number = 5
        A = int(Z * a_multiplier)
        zaid = Z * 1000 + A
        nuclide = Nuclide(zaid)
        mat.append((nuclide, fraction))
        assert mat[0][0] == nuclide
        assert mat[0][1] == pytest.approx(fraction)
        _.verify_export(mat)

    def test_material_append_bad(_):
        mat = Material()
        with pytest.raises(TypeError):
            mat.append(5)
        with pytest.raises(ValueError):
            mat.append((1, 2, 3))
        with pytest.raises(TypeError):
            mat.append(("hi", 1))
        with pytest.raises(TypeError):
            mat.append((Nuclide("1001.80c"), "hi"))
        with pytest.raises(ValueError):
            mat.append((Nuclide("1001.80c"), -1.0))

    @pytest.mark.parametrize(
        "content, strict, is_in",
        [
            ("1001.80c", False, True),
            (1001, False, True),
            ("H-1", False, True),
            (Element(1), False, True),
            (Nucleus(Element(1), 1), False, True),
            (Element(43), False, False),
            ("B-10.00c", False, False),
            ("H", False, True),
            ("H", True, False),
            (Nucleus(Element(5), 10), False, False),
            ("Ni", False, True),
            ("Ni", True, False),  # test wrong library
            ("Ni-0.60c", True, True),
            ("Co-60", False, False),
            ("Co-60", True, False),
            ("Co-60m2.50c", True, True),
        ],
    )
    def test_material_contains(_, materials, content, strict, is_in):
        for big_material in materials:
            if not strict:
                assert is_in == (
                    content in big_material
                ), "Contains didn't work properly"
            assert is_in == big_material.contains_all(content, strict=strict)
            assert is_in == big_material.contains_any(content, strict=strict)
            with pytest.raises(TypeError):
                {} in big_material
            with pytest.raises(TypeError):
                big_material.contains_all("H", strict=5)
            with pytest.raises(TypeError):
                big_material.contains_any("H", strict=5)

    def test_material_multi_contains(_, materials):
        for big_material in materials:
            # contains all
            assert big_material.contains_all("1001", "U-235", "Pu-239", threshold=0.01)
            assert not big_material.contains_all(
                "1001", "U-235", "Pu-239", threshold=0.01, strict=True
            )
            assert not big_material.contains_all(
                "1001", "U-235", "Pu-239", threshold=0.07
            )
            assert not big_material.contains_all("U-235", "B-10")
            # contains any
            assert not big_material.contains_any("C", "B", "F")
            print("sadness")
        assert big_material.contains_any("h-1", "C", "B")

    def test_material_contains_bad(_):
        mat = Material()
        for method in [mat.contains_all, mat.contains_any]:
            with pytest.raises(TypeError):
                method(mat)
            with pytest.raises(TypeError):
                method("1001", mat)
            with pytest.raises(ValueError):
                method("hi")
            with pytest.raises(TypeError):
                method("1001", threshold="hi")
            with pytest.raises(ValueError):
                method("1001", threshold=-1.0)

    def test_material_normalize(_, materials):
        for big_material in materials:
            # make sure it's not an invalid starting condition
            assert sum(big_material.values) != pytest.approx(1.0)
            answer = 1.0 / len(big_material)
            big_material.normalize()
            for value in big_material.values:
                assert value == pytest.approx(answer)

    @pytest.mark.parametrize(
        "kwargs, length",
        [
            ({"name": "H"}, 6),
            ({"name": "H-1"}, 4),
            ({"name": "H-1.04c"}, 1),
            ({"name": "H-1.00c"}, 1),
            ({"name": "U235m1"}, 1),
            ({"element": Element(1)}, 6),
            ({"element": "H"}, 6),
            ({"element": "H", "strict": True}, 0),
            ({"element": "H", "A": 1, "strict": True}, 0),
            (
                {
                    "element": "H",
                    "A": 1,
                    "library": slice("00c", "05c"),
                    "strict": True,
                },
                2,
            ),
            ({"element": slice(92, 95)}, 5),
            ({"A": 1}, 4),
            ({"A": slice(235, 240)}, 5),
            ({"A": slice(232, 243, 2)}, 5),
            ({"A": slice(None)}, 17),
            ({"meta_state": 0}, 14),
            ({"meta_state": 1}, 2),
            ({"meta_state": slice(0, 2)}, 16),
            ({"library": "80c"}, 3),
            ({"library": slice("00c", "10c")}, 2),
        ],
    )
    def test_material_find(_, materials, kwargs, length):
        for big_material in materials:
            returned = list(big_material.find(**kwargs))
            assert len(returned) == length
            for idx, (nuclide, fraction) in returned:
                assert isinstance(idx, int)
                assert isinstance(nuclide, Nuclide)
                assert isinstance(fraction, float)
            returned = list(big_material.find_vals(**kwargs))
            assert len(returned) == length
            for fraction in returned:
                assert isinstance(fraction, float)

    def test_material_find_bad(_, materials):
        for big_material in materials:
            with pytest.raises(TypeError):
                list(big_material.find(_))
            with pytest.raises(ValueError):
                list(big_material.find("not_good"))
            with pytest.raises(TypeError):
                list(big_material.find(A="hi"))
            with pytest.raises(TypeError):
                list(big_material.find(meta_state="hi"))
            with pytest.raises(TypeError):
                list(big_material.find(element=1.23))
            with pytest.raises(TypeError):
                list(big_material.find(library=5))
            with pytest.raises(TypeError):
                list(big_material.find(strict=5))

    def test_material_str(_):
        in_str = "M20 1001.80c 0.5 8016.80c 0.4 94239.80c 0.1"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        answers = """\
MATERIAL: 20 fractions: atom
 H-1     (80c) 0.5
 O-16    (80c) 0.4
Pu-239   (80c) 0.1
"""
        output = repr(material)
        print(output)
        assert output == answers
        output = str(material)
        print(output)
        assert output == "MATERIAL: 20, ['hydrogen', 'oxygen', 'plutonium']"

    def test_material_sort(_):
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material1 = Material(input_card)
        in_str = "M30 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material2 = Material(input_card)
        sort_list = sorted([material2, material1])
        answers = [material1, material2]
        for i, mat in enumerate(sort_list):
            assert mat == answers[i]

    def test_material_format_mcnp(_):
        in_strs = ["M20 1001.80c 0.5", "     8016.80c         0.5"]
        input_card = Input(in_strs, BlockType.DATA)
        material = Material(input_card)
        material.number = 25
        answers = ["M25 1001.80c 0.5", "     8016.80c         0.5"]
        output = material.format_for_mcnp_input((6, 2, 0))
        assert output == answers

    def test_material_comp_init(_):
        with pytest.raises(DeprecationWarning):
            MaterialComponent(Nuclide("1001"), 0.1)

    def test_mat_comp_init_warn(_):
        with pytest.raises(DeprecationWarning):
            MaterialComponent(Nuclide("1001.80c"), 0.1)

    def test_mat_eq(_, materials):
        for big_material in materials:
            new_mat = big_material.clone()
            new_mat.number = big_material.number
            assert new_mat == big_material
            assert new_mat != 5
            new_mat.values[-1] += 1.5
            assert new_mat != big_material
            new_mat.nuclides[-1].library = "09c"
            assert new_mat != big_material
            del new_mat[0]
            assert new_mat != big_material
            new_mat.number = 23
            assert new_mat != big_material

    def test_mat_long_str(_, materials):
        for big_material in materials:
            for i in range(23, 30):
                big_material.add_nuclide(Nuclide(element=Element(i)), 0.123)
            str(big_material)
            repr(big_material)

    @pytest.mark.parametrize("file", ["test.imcnp", "pin_cell.imcnp"])
    def test_read_add_write(_, file):
        problem = montepy.read_input(pathlib.Path("tests") / "inputs" / file)
        mat = problem.materials[2]
        mat.add_nuclide("O-16.80c", 0.3)
        try:
            out_file = f"mat_read_write_21ij43werr{file}"
            problem.write_problem(out_file)
            with open(out_file, "r") as fh:
                for line in fh:
                    print(line.rstrip())
            new_problem = montepy.read_input(out_file)
        finally:
            pathlib.Path(out_file).unlink(True)

    @pytest.mark.parametrize(
        "line, mat_number, is_atom, fractions",
        [
            ("M20 1001.80c 0.5 8016.710nc 0.5", 20, True, [0.5, 0.5]),
            ("m1 1001 0.33 8016 0.666667", 1, True, [0.33, 0.666667]),
            ("M20 1001 0.5 8016 0.5", 20, True, [0.5, 0.5]),
            ("M20 1001.80c -0.5 8016.80c -0.5", 20, False, [0.5, 0.5]),
            ("M20 1001.80c -0.5 8016.710nc -0.5", 20, False, [0.5, 0.5]),
            ("M20 1001.80c 0.5 8016.80c 0.5 Gas=1", 20, True, [0.5, 0.5]),
            (
                "m1      8016.71c  2.6999999-02 8017.71c  9.9999998-01 plib=84p",
                1,
                True,
                [2.6999999e-2, 9.9999998e-01],
            ),
            *[
                (f"M20 1001.80c 0.5 8016.80c 0.5 {part}={lib}", 20, True, [0.5, 0.5])
                for part, lib in [
                    ("nlib", "80c"),
                    ("nlib", "701nc"),
                    ("estep", 1),
                    ("pnlib", "710nc"),
                    ("slib", "80c"),
                ]
            ],
        ],
    )
    def test_material_init(_, line, mat_number, is_atom, fractions):
        input = Input([line], BlockType.DATA)
        material = Material(input)
        assert material.number == mat_number
        assert material.old_number == mat_number
        assert material.is_atom_fraction == is_atom
        for component, gold in zip(material, fractions):
            assert component[1] == pytest.approx(gold)
        if "gas" in line:
            assert material.parameters["gas"]["data"][0].value == pytest.approx(1.0)

    @pytest.mark.parametrize(
        "line", ["Mfoo", "M-20", "M20 1001.80c foo", "M20 1001.80c 0.5 8016.80c -0.5"]
    )
    def test_bad_init(_, line):
        # test invalid material number
        input = Input([line], BlockType.DATA)
        with pytest.raises(MalformedInputError):
            Material(input)

    @pytest.mark.filterwarnings("ignore")
    @given(st.integers(), st.integers())
    def test_mat_clone(_, start_num, step):
        input = Input(["m1 1001.80c 0.3 8016.80c 0.67"], BlockType.DATA)
        mat = Material(input)
        problem = montepy.MCNP_Problem("foo")
        for prob in {None, problem}:
            mat.link_to_problem(prob)
            if prob is not None:
                problem.materials.append(mat)
            if start_num <= 0 or step <= 0:
                with pytest.raises(ValueError):
                    mat.clone(start_num, step)
                return
            new_mat = mat.clone(start_num, step)
            assert new_mat is not mat
            for (iso, fraction), (gold_iso, gold_fraction) in zip(new_mat, mat):
                assert iso is not gold_iso
                assert iso.ZAID == gold_iso.ZAID
                assert fraction == pytest.approx(gold_fraction)
            assert new_mat._number is new_mat._tree["classifier"].number
            output = new_mat.format_for_mcnp_input((6, 3, 0))
            input = Input(output, BlockType.DATA)
            newer_mat = Material(input)
            assert newer_mat.number == new_mat.number

    @pytest.mark.parametrize(
        "args, error",
        [
            (("c", 1), TypeError),
            ((1, "d"), TypeError),
            ((-1, 1), ValueError),
            ((0, 1), ValueError),
            ((1, 0), ValueError),
            ((1, -1), ValueError),
        ],
    )
    def test_mat_clone_bad(_, args, error):
        input = Input(["m1 1001.80c 0.3 8016.80c 0.67"], BlockType.CELL)
        mat = Material(input)
        with pytest.raises(error):
            mat.clone(*args)

    def test_mat_num_init(_):
        mat = Material(number=5)
        assert mat.number == 5

    @settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
    @given(
        lib_num=st.integers(0, 99),
        extra_char=st.characters(min_codepoint=97, max_codepoint=122),
        lib_suffix=st.sampled_from("cdmgpuyehorsa"),
    )
    def test_mat_change_lib(_, materials, lib_num, extra_char, lib_suffix):
        for big_material in materials:
            mat = big_material.clone()
            library = f"{lib_num:02g}"
            if lib_num >= 100:
                library += extra_char
            library += lib_suffix
            for wrapper in {str, Library}:
                mat.change_libraries(wrapper(library))
                for nuclide in mat.nuclides:
                    assert nuclide.library == Library(library)
            _.verify_export(mat)

    def test_mat_change_lib_bad(_):
        mat = Material()
        with pytest.raises(TypeError):
            mat.change_libraries(5)
        with pytest.raises(ValueError):
            mat.change_libraries("hi")

    @given(st.integers(1, 99), st.floats(1.9, 2.3), st.floats(0, 20, allow_nan=False))
    def test_mat_add_nuclide(_, Z, a_multiplier, fraction):
        mat = montepy.Material()
        A = int(Z * a_multiplier)
        ZAID = Z * 1000 + A
        for wrapper in {str, Nuclide}:
            mat.add_nuclide(wrapper(ZAID), fraction)
            assert mat.nuclides[-1].ZAID == ZAID
            assert mat.values[-1] == fraction
        with pytest.raises(TypeError):
            mat.add_nuclide(5.0, 5.0)
        with pytest.raises(TypeError):
            mat.add_nuclide(Nuclide("1001.80c"), "hi")
        with pytest.raises(ValueError):
            mat.add_nuclide(Nuclide("1001.80c"), -1.0)

    @pytest.mark.filterwarnings("ignore::montepy.errors.LineExpansionWarning")
    def test_add_nuclide_export(_, materials):
        for big_material in materials:
            _.verify_export(big_material)

    def verify_export(_, mat):
        output = mat.format_for_mcnp_input((6, 3, 0))
        print("Material output", output)
        new_mat = Material(Input(output, BlockType.DATA))
        assert mat.number == new_mat.number, "Material number not preserved."
        assert len(mat) == len(new_mat), "number of components not kept."
        assert mat.is_atom_fraction == new_mat.is_atom_fraction
        for (old_nuc, old_frac), (new_nuc, new_frac) in zip(mat, new_mat):
            assert old_nuc == new_nuc, "Material didn't preserve nuclides."
            assert old_frac == pytest.approx(new_frac)
        for (old_type, old_lib), (new_type, new_lib) in zip(
            mat.default_libraries.items(), new_mat.default_libraries.items()
        ):
            assert old_type == new_type
            assert old_lib == new_lib


class TestThermalScattering:
    def test_thermal_scattering_init(_):
        # test wrong input type assertion
        input_card = Input(["M20"], BlockType.DATA)
        with pytest.raises(MalformedInputError):
            ThermalScatteringLaw(input_card)

        input_card = Input(["Mt20 grph.20t"], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        assert card.old_number == 20
        assert card.thermal_scattering_laws == ["grph.20t"]

        input_card = Input(["Mtfoo"], BlockType.DATA)
        with pytest.raises(MalformedInputError):
            ThermalScatteringLaw(input_card)
        input_card = Input(["Mt-20"], BlockType.DATA)
        with pytest.raises(MalformedInputError):
            ThermalScatteringLaw(input_card)
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        card = ThermalScatteringLaw(material=material)
        assert card.parent_material == material

    def test_thermal_scattering_particle_parser(_):
        # replicate issue #121
        input_card = Input(["Mt20 h-h2o.40t"], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        assert card.old_number == 20
        assert card.thermal_scattering_laws == ["h-h2o.40t"]

    def test_thermal_scatter_validate(_):
        thermal = ThermalScatteringLaw()
        with pytest.raises(montepy.errors.IllegalState):
            thermal.validate()
        with pytest.raises(montepy.errors.IllegalState):
            thermal.format_for_mcnp_input((6, 2, 0))
        material = Material()
        material.number = 1
        thermal._old_number = montepy.input_parser.syntax_node.ValueNode("1", int)
        thermal.update_pointers([material])
        with pytest.raises(montepy.errors.IllegalState):
            thermal.validate()
        thermal._old_number = montepy.input_parser.syntax_node.ValueNode("2", int)
        with pytest.raises(montepy.errors.MalformedInputError):
            thermal.update_pointers([material])
        with pytest.raises(montepy.errors.IllegalState):
            thermal.validate()
        thermal._old_number = montepy.input_parser.syntax_node.ValueNode("2", int)
        with pytest.raises(montepy.errors.MalformedInputError):
            thermal.update_pointers([material])

    def test_thermal_scattering_add(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        card.add_scattering_law("grph.21t")
        self.assertEqual(len(card.thermal_scattering_laws), 2)
        self.assertEqual(card.thermal_scattering_laws, ["grph.20t", "grph.21t"])
        card.thermal_scattering_laws = ["grph.22t"]
        self.assertEqual(card.thermal_scattering_laws, ["grph.22t"])

    def test_thermal_scattering_setter(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        laws = ["grph.21t"]
        card.thermal_scattering_laws = laws
        self.assertEqual(card.thermal_scattering_laws, laws)
        with self.assertRaises(TypeError):
            card.thermal_scattering_laws = 5
        with self.assertRaises(TypeError):
            card.thermal_scattering_laws = [5]

    def test_thermal_scattering_material_add(self):
        in_str = "M20 1001.80c 1.0"
        input_card = Input([in_str], BlockType.DATA)
        card = Material(input_card)
        card.add_thermal_scattering("grph.21t")
        self.assertEqual(len(card.thermal_scattering.thermal_scattering_laws), 1)
        self.assertEqual(card.thermal_scattering.thermal_scattering_laws, ["grph.21t"])
        card.thermal_scattering_laws = ["grph.22t"]
        self.assertEqual(card.thermal_scattering_laws, ["grph.22t"])
        with self.assertRaises(TypeError):
            card.add_thermal_scattering(5)

    def test_thermal_scattering_format_mcnp(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        material.thermal_scattering = card
        card._parent_material = material
        material.thermal_scattering.thermal_scattering_laws = ["grph.20t"]
        self.assertEqual(card.format_for_mcnp_input((6, 2, 0)), ["Mt20 grph.20t "])

    def test_thermal_str(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        self.assertEqual(str(card), "THERMAL SCATTER: ['grph.20t']")
        self.assertEqual(
            repr(card),
            "THERMAL SCATTER: material: None, old_num: 20, scatter: ['grph.20t']",
        )

    def test_thermal_scattering_add(_):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        card.add_scattering_law("grph.21t")
        assert len(card.thermal_scattering_laws) == 2
        assert card.thermal_scattering_laws == ["grph.20t", "grph.21t"]
        card.thermal_scattering_laws = ["grph.22t"]
        assert card.thermal_scattering_laws == ["grph.22t"]

    def test_thermal_scattering_setter(_):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        laws = ["grph.21t"]
        card.thermal_scattering_laws = laws
        assert card.thermal_scattering_laws == laws
        with pytest.raises(TypeError):
            card.thermal_scattering_laws = 5
        with pytest.raises(TypeError):
            card.thermal_scattering_laws = [5]

    def test_thermal_scattering_material_add(_):
        in_str = "M20 1001.80c 1.0"
        input_card = Input([in_str], BlockType.DATA)
        card = Material(input_card)
        card.add_thermal_scattering("grph.21t")
        assert len(card.thermal_scattering.thermal_scattering_laws) == 1
        assert card.thermal_scattering.thermal_scattering_laws == ["grph.21t"]
        card.thermal_scattering.thermal_scattering_laws = ["grph.22t"]
        assert card.thermal_scattering.thermal_scattering_laws == ["grph.22t"]
        with pytest.raises(TypeError):
            card.add_thermal_scattering(5)

    def test_thermal_scattering_format_mcnp(_):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        material.thermal_scattering = card
        card._parent_material = material
        material.thermal_scattering.thermal_scattering_laws = ["grph.20t"]
        card.format_for_mcnp_input((6, 2, 0)) == ["Mt20 grph.20t "]

    def test_thermal_str(_):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        assert str(card) == "THERMAL SCATTER: ['grph.20t']"
        assert (
            repr(card)
            == "THERMAL SCATTER: material: None, old_num: 20, scatter: ['grph.20t']"
        )


class TestDefaultLib:

    @pytest.fixture
    def mat(_):
        mat = Material()
        mat.number = 1
        return mat

    @pytest.fixture
    def dl(_, mat):
        return DL(mat)

    def test_dl_init(_, dl):
        assert isinstance(dl._parent(), Material)
        assert isinstance(dl._libraries, dict)

    @pytest.mark.parametrize(
        "lib_type, lib", [("nlib", "80c"), ("plib", "80p"), ("alib", "24a")]
    )
    def test_set_get(_, dl, lib_type, lib):
        lib_type_load = LibraryType(lib_type.upper())
        dl[lib_type] = lib
        assert dl[lib_type] == Library(lib), "Library not properly stored."
        assert (
            len(dl._parent()._tree["data"]) == 1
        ), "library not added to parent material"
        dl[lib_type_load] = Library(lib)
        dl[lib_type_load] == Library(lib), "Library not properly stored."
        del dl[lib_type]
        assert (
            len(dl._parent()._tree["data"]) == 0
        ), "library not deleted from parent material"
        assert dl[lib_type] is None, "Default libraries did not delete"
        assert dl["hlib"] is None, "Default value not set."

    def test_bad_set_get(_, dl):
        with pytest.raises(TypeError):
            dl[5] = "80c"
        with pytest.raises(TypeError):
            dl["nlib"] = 5
        with pytest.raises(TypeError):
            del dl[5]

    def test_dl_str(_, dl):
        str(dl)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
import montepy
from montepy.numbered_object_collection import NumberedObjectCollection
from montepy.universe import Universe


class Universes(NumberedObjectCollection):
    """A container of multiple :class:`~montepy.universe.Universe` instances.

    Notes
    -----

    For examples see the ``NumberedObjectCollection`` :ref:`collect ex`.
    """

    def __init__(self, objects: list = None, problem: montepy.MCNP_Problem = None):
        super().__init__(Universe, objects, problem)



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.

from __future__ import annotations
import collections as co
import copy
from typing import Generator, Union
from numbers import Integral, Real

import montepy
from montepy.numbered_object_collection import NumberedDataObjectCollection

Material = montepy.data_inputs.material.Material


class Materials(NumberedDataObjectCollection):
    """A container of multiple :class:`~montepy.data_inputs.material.Material` instances.

    Notes
    -----
    When items are added to this (and this object is linked to a problem),
    they will also be added to :func:`montepy.mcnp_problem.MCNP_Problem.data_inputs`.

    Notes
    -----

    For examples see the ``NumberedObjectCollection`` :ref:`collect ex`.

    Parameters
    ----------
    objects : list
        the list of materials to start with if needed
    """

    def __init__(self, objects=None, problem=None):
        super().__init__(Material, objects, problem)

    def get_containing_any(
        self,
        *nuclides: Union[
            montepy.data_inputs.nuclide.Nuclide,
            montepy.data_inputs.nuclide.Nucleus,
            montepy.Element,
            str,
            int,
        ],
        threshold: float = 0.0,
        strict: bool = False,
    ) -> Generator[Material]:
        """Get all materials that contain any of these these nuclides.

        This uses :func:`~montepy.data_inputs.material.Material.contains` under the hood.
        See that documentation for more guidance.

        Examples
        ^^^^^^^^

        One example would to be find all water bearing materials:

        .. testcode::

            import montepy
            problem = montepy.read_input("foo.imcnp")
            for mat in problem.materials.get_containing_any("H-1", "U-235", threshold = 0.3):
                print(mat)

        .. testoutput::

            MATERIAL: 1, ['hydrogen', 'oxygen']

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *nuclides : Union[Nuclide, Nucleus, Element, str, int]
            a plurality of nuclides to check for.
        threshold : float
            the minimum concentration of a nuclide to be considered. The
            material components are not first normalized.
        strict : bool
            If True this does not let an elemental nuclide match all
            child isotopes, isomers, nor will an isotope match all
            isomers, nor will a blank library match all libraries.

        Returns
        -------
        Generator[Material]
            A generator of all matching materials

        Raises
        ------
        TypeError
            if any argument is of the wrong type.
        ValueError
            if the fraction is not positive or zero, or if nuclide
            cannot be interpreted as a Nuclide.
        """
        return self._contains_arb(
            *nuclides, bool_func=any, threshold=threshold, strict=strict
        )

    def get_containing_all(
        self,
        *nuclides: Union[
            montepy.data_inputs.nuclide.Nuclide,
            montepy.data_inputs.nuclide.Nucleus,
            montepy.Element,
            str,
            int,
        ],
        threshold: float = 0.0,
        strict: bool = False,
    ) -> Generator[Material]:
        """Get all materials that contain all of these nuclides.

        This uses :func:`~montepy.data_inputs.material.Material.contains` under the hood.
        See that documentation for more guidance.

        Examples
        ^^^^^^^^

        One example would to be find all water bearing materials:

        .. testcode::

            import montepy
            problem = montepy.read_input("foo.imcnp")
            for mat in problem.materials.get_containing_all("H-1", "O-16", threshold = 0.3):
                print(mat)

        .. testoutput::

            MATERIAL: 1, ['hydrogen', 'oxygen']

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *nuclides : Union[Nuclide, Nucleus, Element, str, int]
            a plurality of nuclides to check for.
        threshold : float
            the minimum concentration of a nuclide to be considered. The
            material components are not first normalized.
        strict : bool
            If True this does not let an elemental nuclide match all
            child isotopes, isomers, nor will an isotope match all
            isomers, nor will a blank library match all libraries.

        Returns
        -------
        Generator[Material]
            A generator of all matching materials

        Raises
        ------
        TypeError
            if any argument is of the wrong type.
        ValueError
            if the fraction is not positive or zero, or if nuclide
            cannot be interpreted as a Nuclide.
        """
        return self._contains_arb(
            *nuclides, bool_func=all, threshold=threshold, strict=strict
        )

    def _contains_arb(
        self,
        *nuclides: Union[
            montepy.data_inputs.nuclide.Nuclide,
            montepy.data_inputs.nuclide.Nucleus,
            montepy.Element,
            str,
            int,
        ],
        bool_func: co.abc.Callable[co.abc.Iterable[bool]],
        threshold: float = 0.0,
        strict: bool = False,
    ) -> Generator[Material]:
        nuclide_finders = []
        for nuclide in nuclides:
            nuclide_finders.append(Material._promote_nuclide(nuclide, strict))

        def sort_by_type(nuclide):
            type_map = {
                montepy.data_inputs.element.Element: 0,
                montepy.data_inputs.nuclide.Nucleus: 1,
                montepy.data_inputs.nuclide.Nuclide: 2,
            }
            return type_map[type(nuclide)]

        # optimize by most hashable and fail fast
        nuclide_finders = sorted(nuclide_finders, key=sort_by_type)
        for material in self:
            if material._contains_arb(
                *nuclide_finders,
                bool_func=bool_func,
                threshold=threshold,
                strict=strict,
            ):
                # maybe? Maybe not?
                # should Materials act like a set?
                yield material

    @property
    def default_libraries(self) -> dict[montepy.LibraryType, montepy.Library]:
        """The default libraries for this problem defined by ``M0``.

        Examples
        ^^^^^^^^

        To set the default libraries for a problem you need to set this dictionary
        to a Library or string.

        .. testcode:: python

            import montepy
            problem = montepy.read_input("foo.imcnp")

            # set neutron default to ENDF/B-VIII.0
            problem.materials.default_libraries["nlib"] = "00c"
            # set photo-atomic
            problem.materials.default_libraries[montepy.LibraryType.PHOTO_ATOMIC] = montepy.Library("80p")

        .. versionadded:: 1.0.0

        Returns
        -------
        dict[LibraryType, Library]
            the default libraries in use
        """
        try:
            return self[0].default_libraries
        except KeyError:
            default = Material()
            default.number = 0
            self.append(default)
            return self.default_libraries

    def mix(
        self,
        materials: list[Material],
        fractions: list[float],
        starting_number=None,
        step=None,
    ) -> Material:
        """Mix the given materials in the provided fractions to create a new material.

        All materials must use the same fraction type, either atom fraction or mass fraction.
        The fractions given to this method are interpreted in that way as well.

        This new material will automatically be added to this collection.

        Examples
        --------

        An example way to mix materials is to first create the materials to mix:

        .. testcode::

            import montepy
            mats = montepy.Materials()
            h2o = montepy.Material()
            h2o.number = 1
            h2o.add_nuclide("1001.80c", 2.0)
            h2o.add_nuclide("8016.80c", 1.0)

            boric_acid = montepy.Material()
            boric_acid.number = 2
            for nuclide, fraction in {
                "1001.80c": 3.0,
                "B-10.80c": 1.0 * 0.189,
                "B-11.80c": 1.0 * 0.796,
                "O-16.80c": 3.0
            }.items():
                boric_acid.add_nuclide(nuclide, fraction)

        Then to make the material mixture you just need to specify the fractions:

        .. testcode::

            boron_ppm = 10
            boric_conc = boron_ppm * 1e-6
            borated_water = mats.mix([h2o, boric_acid], [1 - boric_conc, boric_conc])

        Parameters
        ----------
        materials : list[Material]
            the materials to mix.
        fractions
            the corresponding fractions for each material in either atom
            or mass fractions, depending on the materials fraction type.
        starting_number : Union[int, None]
            the starting number to assign this new material.
        step : Union[int, None]
            the step size to take when finding a new number.

        Returns
        -------
        Material
            a new material with the mixed components of the given
            materials

        Raises
        ------
        TypeError
            if invalid objects are given.
        ValueError
            if the number of elements in the two lists mismatch, or if
            not all the materials are of the same fraction type, or if a
            negative starting_number or step are given.
        """
        if not isinstance(materials, list):
            raise TypeError(f"materials must be a list. {materials} given.")
        if len(materials) == 0:
            raise ValueError(f"materials must be non-empty. {materials} given.")
        for mat in materials:
            if not isinstance(mat, Material):
                raise TypeError(
                    f"material in materials is not of type Material. {mat} given."
                )
            if mat.is_atom_fraction != materials[0].is_atom_fraction:
                raise ValueError(
                    f"All materials must have the same is_atom_fraction value. {mat} is the odd one out."
                )
        if not isinstance(fractions, list):
            raise TypeError(f"fractions must be a list. {fractions} given.")
        for frac in fractions:
            if not isinstance(frac, Real):
                raise TypeError(f"fraction in fractions must be a float. {frac} given.")
            if frac < 0.0:
                raise ValueError(f"Fraction cannot be negative. {frac} given.")
        if len(fractions) != len(materials):
            raise ValueError(
                f"Length of materials and fractions don't match. The lengths are, materials: {len(materials)}, fractions: {len(fractions)}"
            )
        if not isinstance(starting_number, (Integral, type(None))):
            raise TypeError(
                f"starting_number must be an int. {starting_number} of type {type(starting_number)} given."
            )
        if starting_number is not None and starting_number <= 0:
            raise ValueError(
                f"starting_number must be positive. {starting_number} given."
            )
        if not isinstance(step, (Integral, type(None))):
            raise TypeError(f"step must be an int. {step} of type {type(step)} given.")
        if step is not None and step <= 0:
            raise ValueError(f"step must be positive. {step} given.")
        ret = Material()
        if starting_number is None:
            starting_number = self.starting_number
        if step is None:
            step = self.step
        ret.number = self.request_number(starting_number, step)
        ret.is_atom_fraction = materials[0].is_atom_fraction
        new_mats = copy.deepcopy(materials)
        for mat, fraction in zip(new_mats, fractions):
            mat.normalize()
            for nuclide, frac in mat._components:
                frac = copy.deepcopy(frac)
                frac.value *= fraction
                ret._components.append((nuclide, frac))
        return ret



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
import collections as co
import copy
import math
from numbers import Integral, Real
from typing import Generator, Union
import weakref

import montepy
from montepy.data_inputs import data_input, thermal_scattering
from montepy.data_inputs.nuclide import Library, Nucleus, Nuclide
from montepy.data_inputs.element import Element
from montepy.input_parser import syntax_node
from montepy.input_parser.material_parser import MaterialParser
from montepy.numbered_mcnp_object import Numbered_MCNP_Object, InitInput
from montepy.errors import *
from montepy.utilities import *
from montepy.particle import LibraryType


MAX_PRINT_ELEMENTS: int = 5
"""
The maximum number of elements to print in a material string descripton.
"""

DEFAULT_INDENT: int = 6
"""
The default number of spaces to indent on a new line by.

This is used for adding new material components.
By default all components made from scratch are added to their own line with this many leading spaces.
"""

NuclideLike = Union[Nuclide, Nucleus, Element, str, Integral]


class _DefaultLibraries:
    """A dictionary wrapper for handling the default libraries for a material.

    The default libraries are those specified by keyword, e.g., ``nlib=80c``.

    Parameters
    ----------
    parent_mat : Material
        the material that this default library is associated with.
    """

    __slots__ = "_libraries", "_parent"

    def __init__(self, parent_mat: Material):
        self._libraries = {}
        self._parent = weakref.ref(parent_mat)

    def __getitem__(self, key):
        key = self._validate_key(key)
        try:
            return Library(self._libraries[key]["data"].value)
        except KeyError:
            return None

    def __setitem__(self, key, value):
        key = self._validate_key(key)
        if not isinstance(value, (Library, str)):
            raise TypeError("")
        if isinstance(value, str):
            value = Library(value)
        try:
            node = self._libraries[key]
        except KeyError:
            node = self._generate_default_node(key, str(value))
            self._parent()._append_param_lib(node)
            self._libraries[key] = node
        node["data"].value = str(value)

    def __delitem__(self, key):
        key = self._validate_key(key)
        node = self._libraries.pop(key)
        self._parent()._delete_param_lib(node)

    def __str__(self):
        return "\n".join([f"{key} = {value}" for key, value in self.items()])

    def __iter__(self):
        return iter(self._libraries)

    def items(self):
        for lib_type, node in self._libraries.items():
            yield (lib_type, node["data"].value)

    @staticmethod
    def _validate_key(key):
        if not isinstance(key, (str, LibraryType)):
            raise TypeError("")
        if not isinstance(key, LibraryType):
            key = LibraryType(key.upper())
        return key

    @staticmethod
    def _generate_default_node(key: LibraryType, val: str):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = key.value
        ret = {
            "classifier": classifier,
            "seperator": syntax_node.PaddingNode(" = "),
            "data": syntax_node.ValueNode(val, str),
        }
        return syntax_node.SyntaxNode("mat library", ret)

    def _load_node(self, key: Union[str, LibraryType], node: syntax_node.SyntaxNode):
        key = self._validate_key(key)
        self._libraries[key] = node

    def __getstate__(self):
        return {"_libraries": self._libraries}

    def __setstate__(self, state):
        self._libraries = state["_libraries"]

    def _link_to_parent(self, parent_mat: Material):
        self._parent = weakref.ref(parent_mat)


class _MatCompWrapper:
    """A wrapper that allows unwrapping Nuclide and fractions"""

    __slots__ = "_parent", "_index", "_setter"

    def __init__(self, parent, index, setter):
        self._parent = parent
        self._index = index
        self._setter = setter

    def __iter__(self):

        def generator():
            for component in self._parent:
                yield component[self._index]

        return generator()

    def __getitem__(self, idx):
        return self._parent[idx][self._index]

    def __setitem__(self, idx, val):
        new_val = self._setter(self._parent[idx], val)
        self._parent[idx] = new_val


class Material(data_input.DataInputAbstract, Numbered_MCNP_Object):
    """A class to represent an MCNP material.

    Examples
    --------

    First it might be useful to load an example problem:

    .. testcode::

        import montepy
        problem = montepy.read_input("foo.imcnp")
        mat = problem.materials[1]
        print(mat)

    .. testoutput::

        MATERIAL: 1, ['hydrogen', 'oxygen']

    Materials are iterable
    ^^^^^^^^^^^^^^^^^^^^^^

    Materials look like a list of tuples, and is iterable.
    Whether or not the material is defined in mass fraction or atom fraction
    is stored for the whole material in :func:`~montepy.data_inputs.material.Material.is_atom_fraction`.
    The fractions (atom or mass) of the componenets are always positive,
    because MontePy believes in physics.

    .. testcode::

        assert mat.is_atom_fraction # ensures it is in atom_fraction

        for nuclide, fraction in mat:
            print("nuclide", nuclide, fraction)

    This would display:

    .. testoutput::

        nuclide  H-1     (80c) 2.0
        nuclide  O-16    (80c) 1.0

    As a list, Materials can be indexed:

    .. testcode::

        oxygen, ox_frac = mat[1]
        mat[1] = (oxygen, ox_frac + 1e-6)
        del mat[1]

    If you need just the nuclides or just the fractions of components in this material see: :func:`nuclides` and
    :func:`values`.

    You can check if a Nuclide is in a Material
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    You can check if a :class:`~montepy.data_inputs.nuclide.Nuclide` or :class:`~montepy.data_input.element.Element` is
    in a Material with ``in``.

    .. doctest::

        >>> montepy.Nuclide("H-1") in mat
        True
        >>> "H-1" in mat
        True
        >>> montepy.Element(1) in mat
        True
        >>> montepy.Element(92) in mat
        False

    Add New Component
    ^^^^^^^^^^^^^^^^^

    The easiest way to add new components to a material is with
    :func:`~montepy.data_inputs.material.Material.add_nuclide`.

    .. testcode::

        # add boric acid to water
        boric_acid_frac = 1e-6
        mat[0]
        # Add by nuclide object
        mat.add_nuclide(oxygen, ox_frac + 3 * boric_acid_frac)
        # add by nuclide Name or ZAID
        mat.add_nuclide("B-10.80c", 1e-6)
        print(mat)

    .. testoutput::

        MATERIAL: 1, ['hydrogen', 'oxygen', 'boron']

    Default Libraries
    ^^^^^^^^^^^^^^^^^

    Also materials have the concept of :func:`~montepy.data_inputs.material.Material.default_libraries`.
    These are the libraries set by ``NLIB``, ``PLIB``, etc.,
    which are used when a library of the correct :class:`~montepy.particle.LibraryType` is not provided with the
    nuclide.
    :func:`~montepy.data_inputs.material.Material.default_libraries` acts like a dictionary,
    and can accept a string or a :class:`~montepy.particle.LibraryType` as keys.

    .. testcode::

        print(mat.default_libraries["plib"])
        mat.default_libraries[montepy.LibraryType.NEUTRON] = "00c"
        print(mat.default_libraries["nlib"])

    .. testoutput::

        80p
        00c

    See Also
    --------

    * :manual631:`5.6.1`
    * :manual63:`5.6.1`
    * :manual62:`106`


    .. versionchanged:: 1.0.0

        * Added number parameter to constructor.
        * This was the primary change for this release. For more details on what changed see :ref:`migrate 0 1`.
            * Switched to list-like data-structure
            * Added ability to search by Nuclide
            * Added Support for default libraries (e.g., ``nlib=80c``).

    Parameters
    ----------
    input : Union[Input, str]
        The Input syntax object this will wrap and parse.
    number : int
        The number to set for this object.
    """

    _parser = MaterialParser()
    _NEW_LINE_STR = "\n" + " " * DEFAULT_INDENT

    def __init__(
        self,
        input: InitInput = None,
        number: int = None,
    ):
        self._components = []
        self._thermal_scattering = None
        self._is_atom_fraction = True
        self._number = self._generate_default_node(int, -1, None)
        self._number.never_pad = True
        self._elements = set()
        self._nuclei = set()
        self._default_libs = _DefaultLibraries(self)
        super().__init__(input)
        self._load_init_num(number)
        if input:
            num = self._input_number
            self._old_number = copy.deepcopy(num)
            self._number = num
            set_atom_frac = False
            isotope_fractions = self._tree["data"]
            is_first = True
            for group in isotope_fractions:
                if len(group) == 2:
                    self._grab_isotope(*group, is_first=is_first)
                    is_first = False
                else:
                    self._grab_default(*group)
        else:
            self._create_default_tree()

    def _grab_isotope(
        self, nuclide: Nuclide, fraction: syntax_node.ValueNode, is_first: bool = False
    ):
        """Grabs and parses the nuclide and fraction from the init function, and loads it."""
        isotope = Nuclide(node=nuclide)
        fraction.is_negatable_float = True
        if is_first:
            self._is_atom_fraction = not fraction.is_negative
        else:
            # if switching fraction formatting
            if fraction.is_negative == self._is_atom_fraction:
                raise MalformedInputError(
                    input,
                    f"Material definitions for material: {self.number} cannot use atom and mass fraction at the same time",
                )
        self._elements.add(isotope.element)
        self._nuclei.add(isotope.nucleus)
        self._components.append((isotope, fraction))

    def _grab_default(self, param: syntax_node.SyntaxNode):
        """Grabs and parses default libraris from init process."""
        try:
            lib_type = LibraryType(param["classifier"].prefix.value.upper())
            self._default_libs._load_node(lib_type, param)
        # skip extra parameters
        except ValueError:
            pass

    def _create_default_tree(self):
        classifier = syntax_node.ClassifierNode()
        classifier.number = self._number
        classifier.number.never_pad = True
        classifier.prefix = syntax_node.ValueNode("M", str, never_pad=True)
        classifier.padding = syntax_node.PaddingNode(" ")
        mats = syntax_node.MaterialsNode("mat stuff")
        self._tree = syntax_node.SyntaxNode(
            "mats",
            {
                "start_pad": syntax_node.PaddingNode(),
                "classifier": classifier,
                "data": mats,
            },
        )

    def _append_param_lib(self, node: syntax_node.SyntaxNode):
        """Adds the given syntax node to this Material's data list.

        This is called from _DefaultLibraries.
        """
        self._ensure_has_ending_padding()
        self._tree["data"].append_param(node)

    def _delete_param_lib(self, node: syntax_node.SyntaxNode):
        """Deletes the given syntax node from this Material's data list.

        This is called from _DefaultLibraries.
        """
        self._tree["data"].nodes.remove((node,))

    @make_prop_val_node("_old_number")
    def old_number(self) -> int:
        """The material number that was used in the read file

        Returns
        -------
        int
        """
        pass

    @make_prop_pointer("_is_atom_fraction", bool)
    def is_atom_fraction(self) -> bool:
        """If true this constituent is in atom fraction, not weight fraction.

        .. versionchanged:: 1.0.0

            This property is now settable.

        Returns
        -------
        bool
        """
        pass

    @property
    def material_components(self):  # pragma: no cover
        """The internal dictionary containing all the components of this material.

        .. deprecated:: 0.4.1
            MaterialComponent has been deprecated as part of a redesign for the material
            interface due to a critical bug in how MontePy handles duplicate nuclides.
            See :ref:`migrate 0 1`.

        Raises
        ------
        DeprecationWarning
            This has been fully deprecated and cannot be used.
        """
        raise DeprecationWarning(
            f"""material_components is deprecated, and has been removed in MontePy 1.0.0.
See <https://www.montepy.org/migrations/migrate0_1.html> for more information """
        )

    @make_prop_pointer("_default_libs")
    def default_libraries(self):
        """The default libraries that are used when a nuclide doesn't have a relevant library specified.

        Default Libraries
        ^^^^^^^^^^^^^^^^^

        Also materials have the concept of :func:`~montepy.data_inputs.material.Material.default_libraries`.
        These are the libraries set by ``NLIB``, ``PLIB``, etc.,
        which are used when a library of the correct :class:`~montepy.particle.LibraryType` is not provided with the
        nuclide.
        :func:`~montepy.data_inputs.material.Material.default_libraries` acts like a dictionary,
        and can accept a string or a :class:`~montepy.particle.LibraryType` as keys.

        .. testcode::

            print(mat.default_libraries["plib"])
            mat.default_libraries[montepy.LibraryType.NEUTRON] = "00c"
            print(mat.default_libraries["nlib"])

        .. testoutput::

            None
            00c

        .. versionadded:: 1.0.0

        """
        pass

    def get_nuclide_library(
        self, nuclide: Nuclide, library_type: LibraryType
    ) -> Union[Library, None]:
        """Figures out which nuclear data library will be used for the given nuclide in this
        given material in this given problem.

        This follows the MCNP lookup process and returns the first Library to meet these rules.

        #. The library extension for the nuclide. For example if the nuclide is ``1001.80c`` for ``LibraryType("nlib")``, ``Library("80c")`` will be returned.

        #. Next if a relevant nuclide library isn't provided the :func:`~montepy.data_inputs.material.Material.default_libraries` will be used.

        #. Finally if the two other options failed ``M0`` will be checked. These are stored in :func:`montepy.materials.Materials.default_libraries`.

        Notes
        -----

        The final backup is that MCNP will use the first matching library in ``XSDIR``.
        Currently MontePy doesn't support reading an ``XSDIR`` file and so it will return none in this case.


        .. versionadded:: 1.0.0

        Parameters
        ----------
        nuclide : Union[Nuclide, str]
            the nuclide to check.
        library_type : LibraryType
            the LibraryType to check against.

        Returns
        -------
        Union[Library, None]
            the library that will be used in this scenario by MCNP.

        Raises
        ------
        TypeError
            If arguments of the wrong type are given.
        """
        if not isinstance(nuclide, (Nuclide, str)):
            raise TypeError(f"nuclide must be a Nuclide. {nuclide} given.")
        if isinstance(nuclide, str):
            nuclide = Nuclide(nuclide)
        if not isinstance(library_type, (str, LibraryType)):
            raise TypeError(
                f"Library_type must be a LibraryType. {library_type} given."
            )
        if not isinstance(library_type, LibraryType):
            library_type = LibraryType(library_type.upper())
        if nuclide.library.library_type == library_type:
            return nuclide.library
        lib = self.default_libraries[library_type]
        if lib:
            return lib
        if self._problem:
            return self._problem.materials.default_libraries[library_type]
        return None

    def __getitem__(self, idx):
        """"""
        if not isinstance(idx, (Integral, slice)):
            raise TypeError(f"Not a valid index. {idx} given.")
        if isinstance(idx, Integral):
            comp = self._components[idx]
            return self.__unwrap_comp(comp)
        # else it's a slice
        return [self.__unwrap_comp(comp) for comp in self._components[idx]]

    @staticmethod
    def __unwrap_comp(comp):
        return (comp[0], comp[1].value)

    def __iter__(self):
        def gen_wrapper():
            for comp in self._components:
                yield self.__unwrap_comp(comp)

        return gen_wrapper()

    def __setitem__(self, idx, newvalue):
        """"""
        if not isinstance(idx, (Integral, slice)):
            raise TypeError(f"Not a valid index. {idx} given.")
        old_vals = self._components[idx]
        self._check_valid_comp(newvalue)
        node_idx = self._tree["data"].nodes.index((old_vals[0]._tree, old_vals[1]), idx)
        # grab fraction
        old_vals[1].value = newvalue[1]
        self._tree["data"].nodes[node_idx] = (newvalue[0]._tree, old_vals[1])
        self._components[idx] = (newvalue[0], old_vals[1])

    def __len__(self):
        return len(self._components)

    def _check_valid_comp(self, newvalue: tuple[Nuclide, Real]):
        """Checks valid compositions and raises an error if needed."""
        if not isinstance(newvalue, tuple):
            raise TypeError(
                f"Invalid component given. Must be tuple of Nuclide, fraction. {newvalue} given."
            )
        if len(newvalue) != 2:
            raise ValueError(
                f"Invalid component given. Must be tuple of Nuclide, fraction. {newvalue} given."
            )
        if not isinstance(newvalue[0], Nuclide):
            raise TypeError(f"First element must be an Nuclide. {newvalue[0]} given.")
        if not isinstance(newvalue[1], Real):
            raise TypeError(
                f"Second element must be a fraction greater than 0. {newvalue[1]} given."
            )
        if newvalue[1] < 0.0:
            raise ValueError(
                f"Second element must be a fraction greater than 0. {newvalue[1]} given."
            )

    def __delitem__(self, idx):
        if not isinstance(idx, (Integral, slice)):
            raise TypeError(f"Not a valid index. {idx} given.")
        if isinstance(idx, Integral):
            self.__delitem(idx)
            return
        # else it's a slice
        end = idx.start if idx.start is not None else 0
        start = idx.stop if idx.stop is not None else len(self) - 1
        step = -idx.step if idx.step is not None else -1
        for i in range(start, end, step):
            self.__delitem(i)
        if end == 0:
            self.__delitem(0)

    def __delitem(self, idx):
        comp = self._components[idx]
        element = self[idx][0].element
        nucleus = self[idx][0].nucleus
        found_el = False
        found_nuc = False
        # keep indices positive for testing.
        if idx < 0:
            idx += len(self)
        # determine if other components use this element and nucleus
        for i, (nuclide, _) in enumerate(self):
            if i == idx:
                continue
            if nuclide.element == element:
                found_el = True
            if nuclide.nucleus == nucleus:
                found_nuc = True
            if found_el and found_nuc:
                break
        if not found_el:
            self._elements.remove(element)
        if not found_nuc:
            self._nuclei.remove(nucleus)
        self._tree["data"].nodes.remove((comp[0]._tree, comp[1]))
        del self._components[idx]

    def __contains__(self, nuclide):
        if not isinstance(nuclide, (Nuclide, Nucleus, Element, str, Integral)):
            raise TypeError(
                f"Can only check if a Nuclide, Nucleus, Element, or str is in a material. {nuclide} given."
            )
        if isinstance(nuclide, (str, Integral)):
            nuclide = Nuclide(nuclide)
        # switch to elemental
        if isinstance(nuclide, (Nucleus, Nuclide)) and nuclide.A == 0:
            nuclide = nuclide.element
        # switch to nucleus if no library.
        if isinstance(nuclide, Nuclide) and not nuclide.library:
            nuclide = nuclide.nucleus
        if isinstance(nuclide, (Nucleus, Nuclide)):
            if isinstance(nuclide, Nuclide):
                if nuclide.nucleus not in self._nuclei:
                    return False
                for self_nuc, _ in self:
                    if self_nuc == nuclide:
                        return True
                return False
            if isinstance(nuclide, Nucleus):
                return nuclide in self._nuclei
        if isinstance(nuclide, Element):
            element = nuclide
            return element in self._elements

    def append(self, nuclide_frac_pair: tuple[Nuclide, float]):
        """Appends the tuple to this material.

        .. versionadded:: 1.0.0

        Parameters
        ----------
        nuclide_frac_pair : tuple[Nuclide, float]
            a tuple of the nuclide and the fraction to add.
        """
        self._check_valid_comp(nuclide_frac_pair)
        self._elements.add(nuclide_frac_pair[0].element)
        self._nuclei.add(nuclide_frac_pair[0].nucleus)
        # node for fraction
        node = self._generate_default_node(
            float, str(nuclide_frac_pair[1]), self._NEW_LINE_STR
        )
        syntax_node.ValueNode(str(nuclide_frac_pair[1]), float)
        node.is_negatable_float = True
        nuclide_frac_pair = (nuclide_frac_pair[0], node)
        node.is_negative = not self._is_atom_fraction
        self._components.append(nuclide_frac_pair)
        self._ensure_has_ending_padding()
        self._tree["data"].append_nuclide(("_", nuclide_frac_pair[0]._tree, node))

    def _ensure_has_ending_padding(self):
        def get_last_val_node():
            last_vals = self._tree["data"].nodes[-1][-1]
            if isinstance(last_vals, syntax_node.ValueNode):
                return last_vals
            return last_vals["data"]

        if len(self._tree["data"]) == 0:
            return
        padding = get_last_val_node().padding

        def add_new_line_padding():
            if padding is None:
                get_last_val_node().padding = syntax_node.PaddingNode(
                    self._NEW_LINE_STR
                )
            else:
                padding.append(self._NEW_LINE_STR)

        if padding:
            padding.check_for_graveyard_comments(True)
        add_new_line_padding()

    def change_libraries(self, new_library: Union[str, Library]):
        """Change the library for all nuclides in the material.

        .. versionadded:: 1.0.0

        Parameters
        ----------
        new_library : Union[str, Library]
            the new library to set all Nuclides to use.
        """
        if not isinstance(new_library, (Library, str)):
            raise TypeError(
                f"new_library must be a Library or str. {new_library} given."
            )
        if isinstance(new_library, str):
            new_library = Library(new_library)
        for nuclide, _ in self:
            nuclide.library = new_library

    def add_nuclide(self, nuclide: NuclideLike, fraction: float):
        """Add a new component to this material of the given nuclide, and fraction.

        .. versionadded:: 1.0.0

        Parameters
        ----------
        nuclide : Nuclide, str, int
            The nuclide to add, which can be a string Identifier, or
            ZAID.
        fraction : float
            the fraction of this component being added.
        """
        if not isinstance(nuclide, (Nuclide, str, Integral)):
            raise TypeError(
                f"Nuclide must of type Nuclide, str, or int. {nuclide} of type {type(nuclide)} given."
            )
        nuclide = self._promote_nuclide(nuclide, True)
        self.append((nuclide, fraction))

    def contains_all(
        self,
        *nuclides: NuclideLike,
        threshold: float = 0.0,
        strict: bool = False,
    ) -> bool:
        """Checks if this material contains of all of the given nuclides.

        A boolean "and" is used for this comparison.
        That is this material must contain all nuclides at or above the given threshold
        in order to return true.

        Examples
        ^^^^^^^^

        .. testcode::

            import montepy
            problem = montepy.read_input("tests/inputs/test.imcnp")

            # try to find LEU materials
            for mat in problem.materials:
                if mat.contains_all("U-235", threshold=0.02):
                    # your code here
                    pass

            # try to find a uranium
            for mat in problem.materials:
                if mat.contains_all("U"):
                    pass

        Notes
        -----

        The difference between :func:`contains_all` and :func:`contains_any` is only for how they
        handle being given multiple nuclides. This does not impact how given Elements will match
        daughter Nuclides. This is handled instead by ``strict``.

        Notes
        -----

        For details on how to use the ``strict`` argument see the examples in: :func:`find`.


        .. versionadded:: 1.0.0

        Parameters
        ----------
        *nuclides : Union[Nuclide, Nucleus, Element, str, int]
            a plurality of nuclides to check for.
        threshold : float
            the minimum concentration of a nuclide to be considered. The
            material components are not first normalized.
        strict : bool
            If True this does not let an elemental nuclide match all
            child isotopes, isomers, nor will an isotope match all
            isomers, nor will a blank library match all libraries.

        Returns
        -------
        bool
            whether or not this material contains all components given
            above the threshold.

        Raises
        ------
        TypeError
            if any argument is of the wrong type.
        ValueError
            if the fraction is not positive or zero, or if nuclide
            cannot be interpreted as a Nuclide.
        """
        return self._contains_arb(
            *nuclides, bool_func=all, threshold=threshold, strict=strict
        )

    def contains_any(
        self,
        *nuclides: NuclideLike,
        threshold: float = 0.0,
        strict: bool = False,
    ) -> bool:
        """Checks if this material contains any of the given nuclide.

        A boolean "or" is used for this comparison.
        That is, this material must contain any nuclides at or above the given threshold
        in order to return true.

        Examples
        ^^^^^^^^

        .. testcode::

            import montepy
            problem = montepy.read_input("tests/inputs/test.imcnp")

            # try to find any fissile materials
            for mat in problem.materials:
                if mat.contains_any("U-235", "U-233", "Pu-239", threshold=1e-6):
                    pass

        Notes
        -----

        For details on how to use the ``strict`` argument see the examples in: :func:`find`.


        .. versionadded:: 1.0.0

        Parameters
        ----------
        *nuclides : Union[Nuclide, Nucleus, Element, str, int]
            a plurality of nuclides to check for.
        threshold : float
            the minimum concentration of a nuclide to be considered. The
            material components are not first normalized.
        strict : bool
            If True this does not let an elemental nuclide match all
            child isotopes, isomers, nor will an isotope match all
            isomers, nor will a blank library match all libraries.

        Returns
        -------
        bool
            whether or not this material contains all components given
            above the threshold.

        Raises
        ------
        TypeError
            if any argument is of the wrong type.
        ValueError
            if the fraction is not positive or zero, or if nuclide
            cannot be interpreted as a Nuclide.
        """
        return self._contains_arb(
            *nuclides, bool_func=any, threshold=threshold, strict=strict
        )

    @staticmethod
    def _promote_nuclide(nuclide, strict):
        # This is necessary for python 3.9
        if not isinstance(nuclide, (Nuclide, Nucleus, Element, str, Integral)):
            raise TypeError(
                f"Nuclide must be a type that can be converted to a Nuclide. The allowed types are: "
                f"Nuclide, Nucleus, str, int. {nuclide} given."
            )
        if isinstance(nuclide, (str, Integral)):
            nuclide = Nuclide(nuclide)
        # treat elemental as element
        if isinstance(nuclide, (Nucleus, Nuclide)) and nuclide.A == 0 and not strict:
            nuclide = nuclide.element
        if isinstance(nuclide, Nuclide) and not str(nuclide.library) and not strict:
            nuclide = nuclide.nucleus
        return nuclide

    def _contains_arb(
        self,
        *nuclides: Union[Nuclide, Nucleus, Element, str, Integral],
        bool_func: co.abc.Callable[co.abc.Iterable[bool]] = None,
        threshold: float = 0.0,
        strict: bool = False,
    ) -> bool:
        nuclide_finders = []
        if not isinstance(threshold, Real):
            raise TypeError(
                f"Threshold must be a float. {threshold} of type: {type(threshold)} given"
            )
        if threshold < 0.0:
            raise ValueError(f"Threshold must be positive or zero. {threshold} given.")
        if not isinstance(strict, bool):
            raise TypeError(
                f"Strict must be bool. {strict} of type: {type(strict)} given."
            )
        for nuclide in nuclides:
            nuclide_finders.append(self._promote_nuclide(nuclide, strict))

        # fail fast
        if bool_func == all:
            for nuclide in nuclide_finders:
                if nuclide not in self and bool_func == all:
                    return False

        nuclides_search = {}
        nuclei_search = {}
        element_search = {}
        for nuclide in nuclide_finders:
            if isinstance(nuclide, Element):
                element_search[nuclide] = 0.0
            if isinstance(nuclide, Nucleus):
                nuclei_search[nuclide] = 0.0
            if isinstance(nuclide, Nuclide):
                nuclides_search[str(nuclide).lower()] = 0.0

        for nuclide, fraction in self:
            if str(nuclide).lower() in nuclides_search:
                nuclides_search[str(nuclide).lower()] += fraction
            if nuclide.nucleus in nuclei_search:
                nuclei_search[nuclide.nucleus] += fraction
            if nuclide.element in element_search:
                element_search[nuclide.element] += fraction

        threshold_check = lambda x: x > threshold
        return bool_func(
            (
                bool_func(map(threshold_check, nuclides_search.values())),
                bool_func(map(threshold_check, nuclei_search.values())),
                bool_func(map(threshold_check, element_search.values())),
            )
        )

    def clear(self):
        """Clears all nuclide components from this material.

        .. versionadded:: 1.0.0
        """
        for _ in range(len(self)):
            del self[0]

    def normalize(self):
        """Normalizes the components fractions so that they sum to 1.0.

        .. versionadded:: 1.0.0
        """
        total_frac = sum(self.values)
        for _, val_node in self._components:
            val_node.value /= total_frac

    @property
    def values(self):
        """Get just the fractions, or values from this material.

        This acts like a list. It is iterable, and indexable.

        Examples
        ^^^^^^^^

        .. testcode::

            import montepy
            mat = montepy.Material()
            mat.number = 5
            enrichment = 0.04

            # define UO2 with enrichment of 4.0%
            mat.add_nuclide("8016.00c", 2/3)
            mat.add_nuclide("U-235.00c", 1/3 * enrichment)
            mat.add_nuclide("U-238.00c", 2/3 * (1 - enrichment))

            for val in mat.values:
                print(val)
            # iterables can be used with other functions
            max_frac = max(mat.values)
            print("max", max_frac)

        This would print:

        .. testoutput::

            0.6666666666666666
            0.013333333333333332
            0.6399999999999999
            max 0.6666666666666666

        .. testcode::

            # get value by index
            print(mat.values[0])

            # set the value, and double enrichment
            mat.values[1] *= 2.0
            print(mat.values[1])

        This would print:

        .. testoutput::

            0.6666666666666666
            0.026666666666666665

        .. versionadded:: 1.0.0

        Returns
        -------
        Generator[float]
        """

        def setter(old_val, new_val):
            if not isinstance(new_val, Real):
                raise TypeError(
                    f"Value must be set to a float. {new_val} of type {type(new_val)} given."
                )
            if new_val < 0.0:
                raise ValueError(
                    f"Value must be greater than or equal to 0. {new_val} given."
                )
            return (old_val[0], new_val)

        return _MatCompWrapper(self, 1, setter)

    @property
    def nuclides(self):
        """Get just the fractions, or values from this material.

        This acts like a list. It is iterable, and indexable.

        Examples
        ^^^^^^^^

        .. testcode::

            import montepy
            mat = montepy.Material()
            mat.number = 5
            enrichment = 0.04

            # define UO2 with enrichment of 4.0%
            mat.add_nuclide("8016.00c", 2/3)
            mat.add_nuclide("U-235.00c", 1/3 * enrichment)
            mat.add_nuclide("U-238.00c", 2/3 * (1 - enrichment))

            for nuc in mat.nuclides:
                print(repr(nuc))
            # iterables can be used with other functions
            max_zaid = max(mat.nuclides)

        this would print:

        .. testoutput::

            Nuclide('O-16.00c')
            Nuclide('U-235.00c')
            Nuclide('U-238.00c')

        .. testcode::

            # get value by index
            print(repr(mat.nuclides[0]))

            # set the value, and double enrichment
            mat.nuclides[1] = montepy.Nuclide("U-235.80c")

        .. testoutput::

                 Nuclide('O-16.00c')

        .. versionadded:: 1.0.0

        Returns
        -------
        Generator[Nuclide]
        """

        def setter(old_val, new_val):
            if not isinstance(new_val, Nuclide):
                raise TypeError(
                    f"Nuclide must be set to a Nuclide. {new_val} of type {type(new_val)} given."
                )
            return (new_val, old_val[1])

        return _MatCompWrapper(self, 0, setter)

    def __prep_element_filter(self, filter_obj):
        """Makes a filter function for an element.

        For use by find
        """
        if isinstance(filter_obj, str):
            filter_obj = Element.get_by_symbol(filter_obj).Z
        if isinstance(filter_obj, Element):
            filter_obj = filter_obj.Z
        wrapped_filter = self.__prep_filter(filter_obj, "Z")
        return wrapped_filter

    def __prep_filter(self, filter_obj, attr=None):
        """Makes a filter function wrapper"""
        if filter_obj is None:
            return lambda _: True

        if isinstance(filter_obj, slice):

            def slicer(val):
                if attr is not None:
                    val = getattr(val, attr)
                if filter_obj.start:
                    start = filter_obj.start
                    if val < filter_obj.start:
                        return False
                else:
                    start = 0
                if filter_obj.stop:
                    if val >= filter_obj.stop:
                        return False
                if filter_obj.step:
                    if (val - start) % filter_obj.step != 0:
                        return False
                return True

            return slicer
        if attr:
            return lambda val: getattr(val, attr) == filter_obj
        return lambda val: val == filter_obj

    def find(
        self,
        name: str = None,
        element: Union[Element, str, Integral, slice] = None,
        A: Union[int, slice] = None,
        meta_state: Union[int, slice] = None,
        library: Union[str, slice] = None,
        strict: bool = False,
    ) -> Generator[tuple[int, tuple[Nuclide, float]]]:
        """Finds all components that meet the given criteria.

        The criteria are additive, and a component must match all criteria.
        That is the boolean and operator is used.
        Slices can be specified at most levels allowing to search by a range of values.
        For numerical quantities slices are rather intuitive, and follow the same rules that list indices do.
        For elements slices are by Z number only.
        For the library the slicing is done using string comparisons.

        Examples
        --------

        .. testcode::

            import montepy
            mat = montepy.Material()
            mat.number = 1

            # make non-sense material
            for nuclide in ["U-235.80c", "U-238.70c", "Pu-239.00c", "O-16.00c", "C-0", "C-12.00c", "Fe-56"]:
                mat.add_nuclide(nuclide, 0.1)

            print("Get all uranium nuclides.")
            print(list(mat.find(element = "U")))

            print("Get all transuranics")
            print(list(mat.find(element = slice(92, 100))))

            print("Get all ENDF/B-VIII.0")
            print(list(mat.find(library = slice("00c", "09c"))))

        This would print:

        .. testoutput::

            Get all uranium nuclides.
            [(0, (Nuclide('U-235.80c'), 0.1)), (1, (Nuclide('U-238.70c'), 0.1))]
            Get all transuranics
            [(0, (Nuclide('U-235.80c'), 0.1)), (1, (Nuclide('U-238.70c'), 0.1)), (2, (Nuclide('Pu-239.00c'), 0.1))]
            Get all ENDF/B-VIII.0
            [(2, (Nuclide('Pu-239.00c'), 0.1)), (3, (Nuclide('O-16.00c'), 0.1)), (5, (Nuclide('C-12.00c'), 0.1))]

        Strict (Explicit) Matching
        ^^^^^^^^^^^^^^^^^^^^^^^^^^

        Generally this functions treats ambiguity implicitly, and will match as many nuclides as possible.
        This is generally useful, but not always.
        By default when only an element is given all daughter nuclides match, as seen above.
        However, MCNP does provide some "natural" or "elemental" nuclear data,
        and it would be helpful to find these sometimes.
        For instance, you may want to find all instances of elemental nuclides,
        and replace them with explicit isotopes (for instance migrating from ENDF/B-VII.1 to ENDF/B-VIII).
        In these cases the ``strict`` argument is needed.
        When ``strict`` is True an ambiguous ``A`` will only match elemental data:

        .. testcode::

            print("Strict: False", list(mat.find(element="C")))
            print("Strict: True", list(mat.find(element="C", strict=True)))

        will print:

        .. testoutput::

            Strict: False [(4, (Nuclide('C-0'), 0.1)), (5, (Nuclide('C-12.00c'), 0.1))]
            Strict: True [(4, (Nuclide('C-0'), 0.1))]

        Similarly to find nuclides with no library defined you can use strict:

        .. testcode::

            print("Strict: False", list(mat.find(library=None)))
            print("Strict: True", list(mat.find(library=None, strict=True)))

        This would print:

        .. testoutput::

            Strict: False [(0, (Nuclide('U-235.80c'), 0.1)), (1, (Nuclide('U-238.70c'), 0.1)), (2, (Nuclide('Pu-239.00c'), 0.1)), (3, (Nuclide('O-16.00c'), 0.1)), (4, (Nuclide('C-0'), 0.1)), (5, (Nuclide('C-12.00c'), 0.1)), (6, (Nuclide('Fe-56'), 0.1))]
            Strict: True [(4, (Nuclide('C-0'), 0.1)), (6, (Nuclide('Fe-56'), 0.1))]

        .. versionadded:: 1.0.0

        Parameters
        ----------
        name : str
            The name to pass to Nuclide to search by a specific Nuclide.
            If an element name is passed this will only match elemental
            nuclides.
        element : Element, str, int, slice
            the element to filter by, slices must be slices of integers.
            This will match all nuclides that are based on this element.
            e.g., "U" will match U-235 and U-238.
        A : int, slice
            the filter for the nuclide A number.
        meta_state : int, slice
            the metastable isomer filter.
        library : str, slice
            the libraries to limit the search to.
        strict : bool
            When true this will strictly match elements as only elements
            (when no A is given), and only match blank libraries when no
            library is given.

        Returns
        -------
        Generator[tuple[int, tuple[Nuclide, float]]]
            a generator of all matching nuclides, as their index and
            then a tuple of their nuclide, and fraction pairs that
            match.
        """
        # nuclide type enforcement handled by `Nuclide`
        if not isinstance(element, (Element, str, Integral, slice, type(None))):
            raise TypeError(
                f"Element must be only Element, str, int or slice types. {element} of type{type(element)} given."
            )
        if not isinstance(A, (Integral, slice, type(None))):
            raise TypeError(
                f"A must be an int or a slice. {A} of type {type(A)} given."
            )
        if not isinstance(meta_state, (Integral, slice, type(None))):
            raise TypeError(
                f"meta_state must an int or a slice. {meta_state} of type {type(meta_state)} given."
            )
        if not isinstance(library, (str, slice, type(None))):
            raise TypeError(
                f"library must a str or a slice. {library} of type {type(library)} given."
            )
        if not isinstance(strict, bool):
            raise TypeError(
                f"strict must be a bool. {strict} of type {type(strict)} given."
            )
        if name:
            fancy_nuclide = Nuclide(name)
            if fancy_nuclide.A == 0 and not strict:
                element = fancy_nuclide.element
                fancy_nuclide = None
        else:
            fancy_nuclide = None
        if fancy_nuclide and not fancy_nuclide.library:
            first_filter = self.__prep_filter(fancy_nuclide.nucleus, "nucleus")
        else:
            first_filter = self.__prep_filter(fancy_nuclide)

        # create filter for defaults if strict
        if strict:
            # if strict and element switch to A=0
            if element and A is None:
                A = 0
            if library is None:
                library = ""
        filters = [
            first_filter,
            self.__prep_element_filter(element),
            self.__prep_filter(A, "A"),
            self.__prep_filter(meta_state, "meta_state"),
            self.__prep_filter(library, "library"),
        ]
        for idx, component in enumerate(self):
            for filt in filters:
                found = filt(component[0])
                if not found:
                    break
            if found:
                yield idx, component

    def find_vals(
        self,
        name: str = None,
        element: Union[Element, str, int, slice] = None,
        A: Union[int, slice] = None,
        meta_state: Union[int, slice] = None,
        library: Union[str, slice] = None,
        strict: bool = False,
    ) -> Generator[float]:
        """A wrapper for :func:`find` that only returns the fractions of the components.

        For more examples see that function.

        Examples
        ^^^^^^^^

        .. testcode::

            import montepy
            mat = montepy.Material()
            mat.number = 1

            # make non-sense material
            for nuclide in ["U-235.80c", "U-238.70c", "Pu-239.00c", "O-16.00c"]:
                mat.add_nuclide(nuclide, 0.1)

            # get fraction that is uranium
            print(sum(mat.find_vals(element= "U")))

        which would intuitively print:

        .. testoutput::

            0.2

        .. versionadded:: 1.0.0

        Parameters
        ----------
        name : str
            The name to pass to Nuclide to search by a specific Nuclide.
            If an element name is passed this will only match elemental
            nuclides.
        element : Element, str, int, slice
            the element to filter by, slices must be slices of integers.
            This will match all nuclides that are based on this element.
            e.g., "U" will match U-235 and U-238.
        A : int, slice
            the filter for the nuclide A number.
        meta_state : int, slice
            the metastable isomer filter.
        library : str, slice
            the libraries to limit the search to.
        strict : bool
            whether to strictly match elements as only elements (when no
            A is given), and only match blank libraries when no library
            is given.

        Returns
        -------
        Generator[float]
            a generator of fractions whose nuclide matches the criteria.
        """
        for _, (_, fraction) in self.find(
            name, element, A, meta_state, library, strict
        ):
            yield fraction

    def __bool__(self):
        return bool(self._components)

    @make_prop_pointer("_thermal_scattering", thermal_scattering.ThermalScatteringLaw)
    def thermal_scattering(self) -> thermal_scattering.ThermalScatteringLaw:
        """The thermal scattering law for this material

        Returns
        -------
        ThermalScatteringLaw
        """
        return self._thermal_scattering

    @property
    def cells(self) -> Generator[montepy.cell.Cell]:
        """A generator of the cells that use this material.

        Returns
        -------
        Generator[Cell]
            an iterator of the Cell objects which use this.
        """
        if self._problem:
            for cell in self._problem.cells:
                if cell.material == self:
                    yield cell

    def format_for_mcnp_input(self, mcnp_version):
        lines = super().format_for_mcnp_input(mcnp_version)
        if self.thermal_scattering is not None:
            lines += self.thermal_scattering.format_for_mcnp_input(mcnp_version)
        return lines

    def _update_values(self):
        for nuclide, fraction in self._components:
            node = nuclide._tree
            parts = node.value.split(".")
            fraction.is_negative = not self.is_atom_fraction
            if (
                len(parts) > 1
                and parts[-1] != str(nuclide.library)
                or (len(parts) == 1 and str(nuclide.library))
            ):
                node.value = nuclide.mcnp_str()

    def add_thermal_scattering(self, law):
        """Adds thermal scattering law to the material

        Parameters
        ----------
        law : str
            the law that is mcnp formatted
        """
        if not isinstance(law, str):
            raise TypeError(
                f"Thermal Scattering law for material {self.number} must be a string"
            )
        self._thermal_scattering = thermal_scattering.ThermalScatteringLaw(
            material=self
        )
        self._thermal_scattering.add_scattering_law(law)

    def update_pointers(self, data_inputs: list[montepy.data_inputs.DataInput]):
        """Updates pointer to the thermal scattering data

        Parameters
        ----------
        data_inputs : list[DataInput]
            a list of the data inputs in the problem
        """
        pass

    @staticmethod
    def _class_prefix():
        return "m"

    @staticmethod
    def _has_number():
        return True

    @staticmethod
    def _has_classifier():
        return 0

    def __repr__(self):
        ret = f"MATERIAL: {self.number} fractions: "
        if self.is_atom_fraction:
            ret += "atom\n"
        else:
            ret += "mass\n"

        for component in self:
            ret += f"{component[0]} {component[1]}\n"
        if self.thermal_scattering:
            ret += f"Thermal Scattering: {self.thermal_scattering}"

        return ret

    def __str__(self):
        elements = self.get_material_elements()
        print_el = []
        if len(elements) > MAX_PRINT_ELEMENTS:
            print_elements = elements[0:MAX_PRINT_ELEMENTS]
            print_elements.append("...")
            print_elements.append(elements[-1])
        else:
            print_elements = elements
        print_elements = [
            element.name if isinstance(element, Element) else element
            for element in print_elements
        ]
        return f"MATERIAL: {self.number}, {print_elements}"

    def get_material_elements(self):
        """Get the elements that are contained in this material.

        This is sorted by the most common element to the least common.

        Returns
        -------
        list[Element]
            a sorted list of elements by total fraction
        """
        element_frac = co.Counter()
        for nuclide, fraction in self:
            element_frac[nuclide.element] += fraction
        element_sort = sorted(element_frac.items(), key=lambda p: p[1], reverse=True)
        elements = [p[0] for p in element_sort]
        return elements

    def validate(self):
        if len(self._components) == 0 and self.number != 0:
            raise IllegalState(
                f"Material: {self.number} does not have any components defined."
            )

    def __eq__(self, other):
        if not isinstance(other, Material):
            return False
        if self.number != other.number:
            return False
        if len(self) != len(other):
            return False
        my_comp = sorted(self, key=lambda c: c[0])
        other_comp = sorted(other, key=lambda c: c[0])
        for mine, yours in zip(my_comp, other_comp):
            if mine[0] != yours[0]:
                return False
            if not math.isclose(mine[1], yours[1]):
                return False
        return True

    def __setstate__(self, state):
        super().__setstate__(state)
        self._default_libs._link_to_parent(self)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import argparse
import glob
import montepy
from pathlib import Path
import sys

"""
Module to make module executable from CLI.

.. note::
    `__name__ == "__main__"` is unnecessary because this file is not
    run on import.
"""


def define_args(args=None):
    """Sets and parses the command line arguments.

    Returns
    -------
    argparse.NameSpace
        the arguments that were parsed
    """
    parser = argparse.ArgumentParser(
        prog="montepy",
        description="Tool for editing and working with MCNP input files.",
    )
    parser.add_argument(
        "-c",
        "--check",
        action="store",
        nargs="*",
        type=str,
        help="Check the given input file(s) for errors. Accepts globs, and multiple arguments.",
        metavar="input_file",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="Print the version number",
    )
    args = parser.parse_args(args)
    return args


def check_inputs(files):
    """Checks input files for syntax errors.

    Parameters
    ----------
    files : list
        a list of paths to check and show warnings for errors.
    """
    for file in files:
        if not Path(file).is_file():
            raise FileNotFoundError(f"File: {file} not found.")
    for file in files:
        print(f"\n********** Checking: {file} *********\n")
        problem = montepy.MCNP_Problem(file)
        problem.parse_input(True)


def main():  # pragma: no cover
    """The main function"""
    args = define_args()
    if args.check:
        check_inputs(args.check)
    if args.version:
        print(montepy.__version__)


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import unique, Enum


@unique
class Particle(str, Enum):
    """Supported MCNP supported particles.

    Taken from :manual62:`46`.
    """

    NEUTRON = "N"
    PHOTON = "P"
    ELECTRON = "E"
    NEGATIVE_MUON = "|"
    ANTI_NEUTRON = "Q"
    ELECTRON_NEUTRINO = "U"
    MUON_NEUTRINO = "V"
    POSITRON = "F"
    PROTON = "H"
    LAMBDA_BARYON = "L"
    POSITIVE_SIGMA_BARYON = "+"
    NEGATIVE_SIGMA_BARYON = "-"
    CASCADE = "X"
    NEGATIVE_CASCADE = "Y"
    OMEGA_BARYON = "O"
    POSITIVE_MUON = "!"
    ANTI_ELECTRON_NEUTRINO = "<"
    ANTI_MUON_NEUTRINO = ">"
    ANTI_PROTON = "G"
    POSITIVE_PION = "/"
    NEUTRAL_PION = "Z"
    POSITIVE_KAON = "K"
    KAON_SHORT = "%"
    KAON_LONG = "^"
    ANTI_LAMBDA_BARYON = "B"
    ANTI_POSITIVE_SIGMA_BARYON = "_"
    ANTI_NEGATIVE_SIGMA_BARYON = "~"
    ANTI_CASCADE = "C"
    POSITIVE_CASCADE = "W"
    ANTI_OMEGA = "@"
    DEUTERON = "D"
    TRITON = "T"
    HELION = "S"
    ALPHA_PARTICLE = "A"
    NEGATIVE_PION = "*"
    NEGATIVE_KAON = "?"
    HEAVY_ION = "#"

    def __lt__(self, other):
        return self.value < other.value

    def __str__(self):
        return self.name.lower()

    def __eq__(self, other):
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)


@unique
class LibraryType(str, Enum):
    """Enum to represent the possible types that a nuclear data library can be.

    .. versionadded:: 1.0.0

    Taken from :manual63:`5.6.1`.
    """

    def __new__(cls, value, particle=None):
        obj = str.__new__(cls)
        obj._value_ = value
        obj._particle_ = particle
        return obj

    NEUTRON = ("NLIB", Particle.NEUTRON)
    PHOTO_ATOMIC = ("PLIB", None)
    PHOTO_NUCLEAR = ("PNLIB", None)
    ELECTRON = ("ELIB", Particle.ELECTRON)
    PROTON = ("HLIB", Particle.PROTON)
    ALPHA_PARTICLE = ("ALIB", Particle.ALPHA_PARTICLE)
    HELION = ("SLIB", Particle.HELION)
    TRITON = ("TLIB", Particle.TRITON)
    DEUTERON = ("DLIB", Particle.DEUTERON)

    def __str__(self):
        return self.name.lower()

    def __lt__(self, other):
        return self.value < other.value

    def __eq__(self, other):
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase
import pytest

import montepy
from montepy.geometry_operators import Operator
from montepy.input_parser import syntax_node
from montepy.surfaces.half_space import HalfSpace, UnitHalfSpace


def test_halfspace_init():
    surface = montepy.surfaces.CylinderOnAxis()
    surface.number = 1
    node = montepy.input_parser.syntax_node.GeometryTree("hi", {}, "*", " ", " ")
    half_space = HalfSpace(+surface, Operator.UNION, -surface, node)
    assert half_space.operator is Operator.UNION
    assert half_space.left == +surface
    assert half_space.right == -surface
    assert half_space.node is node
    with pytest.raises(TypeError):
        HalfSpace(surface, Operator.UNION)
    with pytest.raises(TypeError):
        HalfSpace(+surface, "hi")
    with pytest.raises(TypeError):
        HalfSpace(+surface, Operator.UNION, surface)
    with pytest.raises(TypeError):
        HalfSpace(+surface, Operator.UNION, -surface, "hi")
    with pytest.raises(ValueError):
        HalfSpace(+surface, Operator.UNION)
    with pytest.raises(ValueError):
        HalfSpace(+surface, Operator.INTERSECTION)


def test_get_leaves():
    surface = montepy.surfaces.CylinderOnAxis()
    surface.number = 1
    cell = montepy.Cell()
    cell.number = 1
    half_space = -surface & ~cell
    cells, surfaces = half_space._get_leaf_objects()
    assert cells == montepy.cells.Cells([cell])
    assert surfaces == montepy.surface_collection.Surfaces([surface])


def test_half_len():
    surface = montepy.surfaces.CylinderOnAxis()
    cell = montepy.Cell()
    surface.number = 1
    cell.number = 1
    half_space = -surface & ~cell
    assert len(half_space) == 2


def test_half_eq():
    cell1 = montepy.Cell()
    cell2 = montepy.Cell()
    cell1.number = 1
    cell2.number = 2
    half1 = ~cell1 & ~cell2
    assert half1 == half1
    half2 = ~cell1 | ~cell2
    assert half1 != half2
    half2 = ~cell1 & ~cell1
    assert half1 != half2
    half2 = ~cell2 & ~cell2
    assert half1 != half2
    half2 = ~half1
    half2._operator = Operator.INTERSECTION
    assert half1 != half2
    with pytest.raises(TypeError):
        half1 == "hi"


def test_half_str():
    cell1 = montepy.Cell()
    cell1.number = 1
    cell2 = montepy.Cell()
    cell2.number = 2
    half_space = ~cell1 | ~cell2
    assert str(half_space) == "(#1:#2)"
    half_space = ~cell1 & ~cell2
    assert str(half_space) == "(#1*#2)"
    half_space = ~(~cell1 & ~cell2)
    assert str(half_space) == "#(#1*#2)"
    # test that no errors occur
    repr(half_space)


def test_unit_half_init():
    node = montepy.input_parser.syntax_node.ValueNode("123", float)
    half_space = UnitHalfSpace(123, True, False, node)
    assert half_space.divider == 123
    assert half_space.side
    assert not half_space._is_cell
    assert half_space.node is node
    with pytest.raises(TypeError):
        UnitHalfSpace("hi", True, False)
    with pytest.raises(TypeError):
        half_space = UnitHalfSpace(123, "hi", False)
    with pytest.raises(TypeError):
        half_space = UnitHalfSpace(123, True, "hi")
    with pytest.raises(TypeError):
        half_space = UnitHalfSpace(123, True, False, "hi")


def test_unit_eq():
    half1 = UnitHalfSpace(1, True, False)
    assert half1 == half1
    half2 = UnitHalfSpace(2, True, False)
    assert half1 != half2
    half2 = UnitHalfSpace(1, False, False)
    assert half1 != half2
    half2 = UnitHalfSpace(1, True, True)
    assert half1 != half2
    with pytest.raises(TypeError):
        half1 == "hi"


def test_unit_str():
    half_space = UnitHalfSpace(1, True, False)
    assert str(half_space) == "+1"
    half_space.side = False
    assert str(half_space) == "-1"
    half_space.is_cell = True
    assert str(half_space), "1"
    cell = montepy.Cell()
    cell.number = 1
    half_space.divider = cell
    assert str(half_space) == "1"


# test geometry integration
def test_surface_half_space():
    surface = montepy.surfaces.cylinder_on_axis.CylinderOnAxis()
    surface.number = 1
    half_space = +surface
    assert isinstance(half_space, HalfSpace)
    assert isinstance(half_space, UnitHalfSpace)
    assert half_space.divider is surface
    assert half_space.side
    half_space = -surface
    assert isinstance(half_space, HalfSpace)
    assert isinstance(half_space, UnitHalfSpace)
    assert half_space.divider is surface
    assert not half_space.side
    with pytest.raises(TypeError):
        half_space = surface & surface
    with pytest.raises(TypeError):
        half_space = surface | surface
    with pytest.raises(TypeError):
        half_space = ~surface


def test_cell_half_space():
    cell = montepy.Cell()
    cell.number = 1
    half_space = ~cell
    assert isinstance(half_space, HalfSpace)
    assert half_space.left.divider is cell
    assert half_space.left.side
    with pytest.raises(TypeError):
        half_space = cell & cell
    with pytest.raises(TypeError):
        half_space = cell | cell
    with pytest.raises(TypeError):
        half_space = +cell
    with pytest.raises(TypeError):
        half_space = -cell


def test_parens_node_export():
    surf1 = montepy.surfaces.cylinder_on_axis.CylinderOnAxis()
    surf2 = montepy.surfaces.cylinder_on_axis.CylinderOnAxis()
    surf3 = montepy.surfaces.cylinder_on_axis.CylinderOnAxis()
    surf1.number = 1
    surf2.number = 2
    surf3.number = 3
    for half_space, form_answer, str_answer in [
        (-surf1 & (+surf2 | -surf3), "-1 (2 : -3)", "(-1*(+2:-3))"),
        ((-surf1 | +surf2) & -surf3, "(-1 : 2) -3", "((-1:+2)*-3)"),
    ]:
        half_space._update_values()
        assert half_space.node.format() == form_answer
        assert str(half_space) == str_answer


def test_intersect_half_space():
    cell1 = montepy.Cell()
    cell2 = montepy.Cell()
    cell1.number = 1
    cell2.number = 2
    cell1 = ~cell1
    cell2 = ~cell2
    half_space = cell1 & cell2
    assert isinstance(half_space, HalfSpace)
    assert half_space.operator is Operator.INTERSECTION
    assert half_space.left is cell1
    assert half_space.right is cell2
    with pytest.raises(TypeError):
        cell1 & "hi"
    with pytest.raises(TypeError):
        "hi" & cell1
    with pytest.raises(TypeError):
        cell2 + cell1
    with pytest.raises(TypeError):
        cell2 - cell1


def test_union_half_space():
    cell1 = montepy.Cell()
    cell2 = montepy.Cell()
    cell1.number = 1
    cell2.number = 2
    cell1 = ~cell1
    cell2 = ~cell2
    half_space = cell1 | cell2
    assert isinstance(half_space, HalfSpace)
    assert half_space.operator is Operator.UNION
    assert half_space.left is cell1
    assert half_space.right is cell2
    with pytest.raises(TypeError):
        cell1 | "hi"
    with pytest.raises(TypeError):
        "hi" | cell1


def test_invert_half_space():
    cell1 = montepy.Cell()
    cell2 = montepy.Cell()
    cell1.number = 1
    cell2.number = 2
    cell1 = ~cell1
    cell2 = ~cell2
    half_space1 = cell1 | cell2
    half_space = ~half_space1
    assert isinstance(half_space, HalfSpace)
    assert half_space.operator is Operator.COMPLEMENT
    assert half_space.left is half_space1
    assert half_space.right is None


def test_iand_recursion():
    cell1 = montepy.Cell()
    cell2 = montepy.Cell()
    cell3 = montepy.Cell()
    cell1.number = 1
    cell2.number = 2
    cell3.number = 3
    half_space = ~cell1 & ~cell2
    cell3.geometry = half_space
    half_space &= ~cell1
    assert half_space.left == ~cell1
    assert not isinstance(half_space.right, UnitHalfSpace)
    assert half_space.right.left == ~cell2
    assert half_space.right.right == ~cell1
    assert half_space.operator == Operator.INTERSECTION
    assert len(half_space) == 3
    for i in range(1, 100):
        half_space &= ~cell1
        assert len(half_space) == i + 3
    with pytest.raises(TypeError):
        half_space &= "hi"
    # test with unit halfspaces
    surf = montepy.surfaces.CylinderParAxis()
    surf.number = 5
    # test going from leaf to tree
    half_space = -surf
    half_space &= +surf
    assert len(half_space) == 2
    # test with actual tree
    half_space = -surf | +surf
    half_space &= +surf
    assert len(half_space) == 3


def test_ior_recursion():
    cell1 = montepy.Cell()
    cell2 = montepy.Cell()
    cell1.number = 1
    cell2.number = 2
    cell1 = ~cell1
    cell2 = ~cell2
    half_space = cell1 | cell2
    half_space |= cell1
    assert half_space.left is cell1
    assert not isinstance(half_space.right, UnitHalfSpace)
    assert half_space.right.left == cell2
    assert half_space.right.right == cell1
    assert half_space.operator == Operator.UNION
    assert len(half_space) == 3
    for i in range(1, 100):
        half_space |= cell1
        assert len(half_space), i + 3
    with pytest.raises(TypeError):
        half_space |= "hi"
    # test with unit halfspaces
    surf = montepy.surfaces.CylinderParAxis()
    surf.number = 5
    half_space = -surf
    half_space |= +surf
    assert len(half_space) == 2
    half_space = -surf | +surf
    half_space |= +surf
    assert len(half_space) == 3


def test_parse_input_tree():
    # simplest intersection
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 1 -2"], montepy.input_parser.block_type.BlockType.CELL
    )
    parser = montepy.input_parser.cell_parser.CellParser()
    tree = parser.parse(input.tokenize())
    geometry = tree["geometry"]
    half_space = HalfSpace.parse_input_node(geometry)
    assert half_space.operator == Operator.INTERSECTION
    assert half_space.left.divider == 1
    assert half_space.left.side
    assert not half_space.left.is_cell
    assert half_space.right.divider == 2
    assert not half_space.right.side
    assert not half_space.right.is_cell
    # Test type checking
    with pytest.raises(TypeError):
        HalfSpace.parse_input_node("hi")
    # test assymetric
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 #2"], montepy.input_parser.block_type.BlockType.CELL
    )
    parser = montepy.input_parser.cell_parser.CellParser()
    tree = parser.parse(input.tokenize())
    geometry = tree["geometry"]
    half_space = HalfSpace.parse_input_node(geometry)
    assert half_space.operator == Operator.COMPLEMENT
    assert half_space.left.divider == 2
    assert half_space.left.side
    assert half_space.left.is_cell
    assert half_space.right is None
    # Test nested trees
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 1 -2 : 3"], montepy.input_parser.block_type.BlockType.CELL
    )
    parser = montepy.input_parser.cell_parser.CellParser()
    tree = parser.parse(input.tokenize())
    geometry = tree["geometry"]
    half_space = HalfSpace.parse_input_node(geometry)
    assert half_space.operator == Operator.UNION
    assert not isinstance(half_space.left, UnitHalfSpace)
    assert isinstance(half_space.right, UnitHalfSpace)
    assert half_space.left.operator == Operator.INTERSECTION
    assert half_space.right.divider == 3
    assert half_space.left.left.divider == 1
    assert half_space.left.right.divider == 2
    # test shift
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 #(1 2 3)"], montepy.input_parser.block_type.BlockType.CELL
    )
    parser = montepy.input_parser.cell_parser.CellParser()
    tree = parser.parse(input.tokenize())
    geometry = tree["geometry"]
    half_space = HalfSpace.parse_input_node(geometry)
    assert half_space.operator == Operator.COMPLEMENT
    assert half_space.right is None
    assert half_space.left.operator == Operator.GROUP
    nested = half_space.left.left
    assert nested.operator == Operator.INTERSECTION
    assert nested.right is not None
    assert nested.left.operator == Operator.INTERSECTION
    # test nested parens
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 (1 : 2:3) 4"], montepy.input_parser.block_type.BlockType.CELL
    )
    parser = montepy.input_parser.cell_parser.CellParser()
    tree = parser.parse(input.tokenize())
    geometry = tree["geometry"]
    half_space = HalfSpace.parse_input_node(geometry)
    assert len(half_space) == 4
    assert half_space.operator == Operator.INTERSECTION
    assert half_space.left.operator == Operator.GROUP
    assert half_space.left.left.operator == Operator.UNION


def test_update_new_parens():
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 1 2"], montepy.input_parser.block_type.BlockType.CELL
    )
    parser = montepy.input_parser.cell_parser.CellParser()
    tree = parser.parse(input.tokenize())
    geometry = tree["geometry"]
    half_space = HalfSpace.parse_input_node(geometry)
    surf3 = montepy.surfaces.cylinder_on_axis.CylinderOnAxis()
    surf3.number = 3
    half2 = half_space.right
    half_space.right = half2 | +surf3
    half_space._update_values()
    assert half_space.node.format() == "1 (2 : 3)"


def test_update_tree_with_comment():
    input = montepy.input_parser.mcnp_input.Input(
        ["1 0 -1 $hi there", "      2"],
        montepy.input_parser.block_type.BlockType.CELL,
    )
    cell = montepy.Cell(input)
    geometry = cell.geometry
    geometry._update_values()


def test_parse_input_value_node():
    node = montepy.input_parser.syntax_node.ValueNode("-1", float)
    half_space = UnitHalfSpace.parse_input_node(node)
    assert not half_space.is_cell
    assert not half_space.side
    assert half_space.divider == 1
    half_space = UnitHalfSpace.parse_input_node(node, True)
    assert half_space.is_cell
    assert half_space.side
    assert half_space.divider == 1
    node = montepy.input_parser.syntax_node.ValueNode("+1", float)
    half_space = UnitHalfSpace.parse_input_node(node, False)
    assert not half_space.is_cell
    assert half_space.side
    assert half_space.divider == 1
    with pytest.raises(TypeError):
        UnitHalfSpace.parse_input_node("hi", False)
    with pytest.raises(TypeError):
        UnitHalfSpace.parse_input_node(node, "hi")


def test_unit_divider_setter():
    cell = montepy.Cell()
    cell.number = 1
    parent = montepy.Cell()
    parent.number = 2
    node = montepy.input_parser.syntax_node.ValueNode("1", float)
    half_space = UnitHalfSpace.parse_input_node(node, True)
    cells = montepy.cells.Cells()
    cells.append(cell)
    half_space.update_pointers(cells, [], parent)
    assert half_space.divider is cell
    cell2 = montepy.Cell()
    # #madLads
    cell2.number = 4
    half_space.divider = cell2
    assert half_space.divider is cell2
    assert cell2 in parent.complements
    # test with surface
    surf = montepy.surfaces.CylinderParAxis()
    surf.number = 5
    with pytest.raises(TypeError):
        half_space.divider = surf
    half_space.is_cell = False
    half_space.divider = surf
    assert half_space.divider is surf
    assert surf in parent.surfaces
    with pytest.raises(TypeError):
        half_space.divider = "hi"


def test_unit_create_default_node():
    surf = montepy.surfaces.CylinderParAxis()
    surf.number = 1
    half_space = UnitHalfSpace(surf, True, False)
    half_space._ensure_has_nodes()
    node = half_space.node
    assert isinstance(node, syntax_node.ValueNode)
    assert node.value == 1
    assert node.format() == "1"
    # test with negative side
    half_space = UnitHalfSpace(surf, False, False)
    half_space._ensure_has_nodes()
    node = half_space.node
    assert node.format() == "-1"
    # test with int and not an object
    half_space = UnitHalfSpace(1, True, False)
    half_space._ensure_has_nodes()
    node = half_space.node
    assert isinstance(node, syntax_node.ValueNode)
    assert node.value == 1


def test_tree_create_default_node():
    surf = montepy.surfaces.CylinderParAxis()
    surf.number = 1
    half_space = -surf | +surf
    half_space._ensure_has_nodes()
    node = half_space.node
    assert isinstance(node, syntax_node.GeometryTree)
    assert node.operator == Operator.UNION
    assert node.format() == "-1 : 1"
    # Test intersection
    half_space = -surf & +surf
    half_space._ensure_has_nodes()
    node = half_space.node
    assert node.operator == Operator.INTERSECTION
    # Test complement
    half_space = ~(-surf | +surf)
    half_space._ensure_has_nodes()
    node = half_space.node
    assert node.operator == Operator.COMPLEMENT
    assert node.format() == " #(-1 : 1)"
    # test cell complement
    cell = montepy.Cell()
    cell.number = 1
    half_space = ~cell
    half_space._ensure_has_nodes()
    node = half_space.node
    assert node.operator == Operator.COMPLEMENT
    assert node.format() == " #1"


def test_update_operators_in_node():
    surf = montepy.surfaces.CylinderParAxis()
    surf.number = 1
    half_space = -surf | +surf
    half_space._ensure_has_nodes()
    half_space.operator = Operator.INTERSECTION
    half_space._update_values()
    assert half_space.node.format() == "-1   1"
    del half_space.right
    half_space.operator = Operator.COMPLEMENT
    cell = montepy.Cell()
    cell.number = 1
    half_space.left = UnitHalfSpace(cell, True, True)
    half_space._update_values()
    assert half_space.node.format() == "  #1"
    # test with blank tree
    half_space = -surf & +surf
    half_space._update_values()
    assert half_space.node.format() == "-1 1"
    # test move to union
    # test centering
    half_space.node.nodes["operator"].append("  ")
    half_space.operator = Operator.UNION
    half_space._update_values()
    assert half_space.node.format() == "-1 : 1"



import pytest

import montepy
from montepy.mcnp_object import MCNP_Object


class TestErrorWrapper:

    def test_error_handler(_):
        obj = ObjectFixture()
        with pytest.raises(ValueError):
            obj.bad_static()
        with pytest.raises(ValueError):
            obj.bad_class()


class ObjectFixture(MCNP_Object):
    def __init__(self):
        pass

    def _update_values(self):
        pass

    @staticmethod
    def bad_static():
        raise ValueError("foo")

    @classmethod
    def bad_class(cls):
        raise ValueError("bar")
